from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple

from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.utils import StrEnum

from webviz_subsurface._models import GruptreeModel
from webviz_subsurface._providers import (
    EnsembleSummaryProvider,
    EnsembleSummaryProviderFactory,
    Frequency,
)

from ._utils import EnsembleGroupTreeData
from ._views._group_tree_view import GroupTreeView, GroupTreeViewElement


class GroupTree(WebvizPluginABC):
    """This plugin vizualizes the network tree and displays pressures,
    rates and other network related information.

    ---
    * **`ensembles`:** Which ensembles in `shared_settings` to include.
    * **`gruptree_file`:** `.csv` with gruptree information.
    * **`time_index`:** Frequency for the data sampling.
    ---

    **Summary data**

    This plugin needs the following summary vectors to be exported:
    * WSTAT for all wells
    * FWIR and FGIR if there are injector wells in the network
    * GOPR, GWPR and GGPR for all group nodes in the production network \
    except the terminal node (GOPRNB etc for BRANPROP trees)
    * GGIR and/or GWIR for all group nodes in the injection network \
    except the terminal node.
    * WOPR, WWPR, WGPR for all producers
    * WWIR and/or WGIR for all injectors

    The following data will be displayed if available:
    * GPR for all group nodes
    * WTHP, WBHP, WMCTL for all wells

    **GRUPTREE input**

    `gruptree_file` is a path to a file stored per realization (e.g. in \
    `share/results/tables/gruptree.csv"`).
    The `gruptree_file` file can be dumped to disk per realization by the `ECL2CSV` forward
    model with subcommand `gruptree`. The forward model uses `ecl2df` to export a table
    representation of the Eclipse network:
    [Link to ecl2csv gruptree documentation.](https://equinor.github.io/ecl2df/usage/gruptree.html).

    **time_index**

    This is the sampling interval of the summary data. It is `yearly` by default, but can be set
    to `monthly` if needed.

    **terminal_node**

    This parameter allows you to specify the terminal node used. It is `FIELD` by default.

    **tree_type**

    This parameter allows you to specify which tree type is vizualised. It is `GRUPTREE` by
    default, but can also be set to `BRANPROP`.

    **excl_well_startswith**

    This parameter allows you to remove wells that starts with any of the strings in this list.
    It is intended to be used to remove f.ex RFT wells that don't have any production or injection.
    Be aware that if actual producers/injectors are removed, the rates in the tree might not be
    consistant.

    **excl_well_endswith**

    Same as excl_well_startswith, but removes wells that ends with any of the strings in this list.

    """

    class Ids(StrEnum):
        GROUPTREE_VIEW = "group-tree-view"

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        ensembles: list,
        gruptree_file: str = "share/results/tables/gruptree.csv",
        rel_file_pattern: str = "share/results/unsmry/*.arrow",
        time_index: str = "yearly",
        terminal_node: str = "FIELD",
        tree_type: str = "GRUPTREE",
        excl_well_startswith: Optional[List] = None,
        excl_well_endswith: Optional[List] = None,
    ) -> None:
        # pylint: disable=too-many-arguments
        super().__init__(stretch=True)

        self._ensembles = ensembles
        self._gruptree_file = gruptree_file

        if excl_well_startswith is None:
            excl_well_startswith = []
        excl_well_startswith = [str(element) for element in excl_well_startswith]
        if excl_well_endswith is None:
            excl_well_endswith = []
        excl_well_endswith = [str(element) for element in excl_well_endswith]

        if ensembles is None:
            raise ValueError('Incorrect argument, must provide "ensembles"')

        self._ensemble_paths: Dict[str, Path] = {
            ensemble_name: webviz_settings.shared_settings["scratch_ensembles"][
                ensemble_name
            ]
            for ensemble_name in ensembles
        }

        provider_factory = EnsembleSummaryProviderFactory.instance()

        self._group_tree_data: Dict[str, EnsembleGroupTreeData] = {}

        sampling = Frequency(time_index)
        for ens_name, ens_path in self._ensemble_paths.items():
            provider: EnsembleSummaryProvider = (
                provider_factory.create_from_arrow_unsmry_presampled(
                    str(ens_path), rel_file_pattern, sampling
                )
            )
            self._group_tree_data[ens_name] = EnsembleGroupTreeData(
                provider=provider,
                gruptree_model=GruptreeModel(
                    ens_name=ens_name,
                    ens_path=ens_path,
                    gruptree_file=gruptree_file,
                    tree_type=tree_type,
                ),
                terminal_node=terminal_node,
                excl_well_startswith=excl_well_startswith,
                excl_well_endswith=excl_well_endswith,
            )

        self.add_view(
            GroupTreeView(self._group_tree_data),
            self.Ids.GROUPTREE_VIEW,
        )

    def add_webvizstore(self) -> List[Tuple[Callable, List[Dict]]]:
        return [
            ens_grouptree_data.webviz_store
            for _, ens_grouptree_data in self._group_tree_data.items()
        ]

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.view(self.Ids.GROUPTREE_VIEW)
                .settings_group(GroupTreeView.Ids.CONTROLS)
                .get_unique_id(),
                "content": "Menu for selecting ensemble and tree mode.",
            },
            {
                "id": self.view(self.Ids.GROUPTREE_VIEW)
                .settings_group(GroupTreeView.Ids.OPTIONS)
                .get_unique_id(),
                "content": "Menu for statistical options or realization.",
            },
            {
                "id": self.view(self.Ids.GROUPTREE_VIEW)
                .settings_group(GroupTreeView.Ids.FILTERS)
                .get_unique_id(),
                "content": "Menu for filtering options.",
            },
            {
                "id": self.view(self.Ids.GROUPTREE_VIEW)
                .view_element(GroupTreeView.Ids.VIEW_ELEMENT)
                .component_unique_id(GroupTreeViewElement.Ids.COMPONENT),
                "content": "Vizualisation of network tree.",
            },
        ]



from typing import Dict, List

from dash import Dash
from webviz_config import WebvizPluginABC, WebvizSettings

from webviz_subsurface._providers.ensemble_grid_provider import (
    EnsembleGridProvider,
    EnsembleGridProviderFactory,
    GridVizService,
)

from ._layout_elements import ElementIds
from ._routes import set_routes
from .views.view_3d._view_3d import View3D


class EXPERIMENTALGridViewerFMU(WebvizPluginABC):
    """
    !> This plugin is an experimental plugin and will see significant changes
    in the future. It is not recommended to use this plugin in a cloud setting.
    There might be unexpected issues and errors in the visualization. Results
    should be validated in ResInsight/RMS!

    The plugin allows for basic visualization of 3D grids and properties
    from FMU ensembles.

    The performance is related to the number of cells,
    and will probably not handle large modelling grids (10+ million cells) at all.

    ---
    * **`ensemble`:** Which ensemble in `shared_settings` to include.

    Provide the grid data either as ROFF:
    * **`roff_grid_name`:** Name of roff grid stored in FMU standard.
    * **`roff_attribute_filter`:** Use an optional subset of roff grid parameters

    Or as Eclipse:
    * **`eclipse_grid_name`:** Case name of the Eclipse files
    * **`eclipse_init_parameters`:** Which Eclipse init parameters to load
    * **`eclipse_restart_parameters`:** Which Eclipse restart parameters to load

    An optional initial cell filter can be set as:
    * **`grid_ijk_filter`:** with one or more of the following:
        ```yaml
            i_min: start column
            i_width: number of columns
            j_min: start row
            j_width: number of rows
            k_min: start layer
            k_width: number of layers
        ```

    ---
    Data can be loaded either from an Eclipse simulation case or from roff files.

    For Eclipse simulations provide the case name as `eclipse_grid_name`, together
    with a list of `eclipse_init_parameters` and `eclipse_restart_parameters`.
    The plugin will then load from the corresponding EGRID, INIT and UNRST files
    in the `eclipse/model` folder.

    For grids stored in ROFF format, the FMU standards are followed. Provide
    a `roff_grid_name` together with an optional `roff_attribute_filter`.
    The corresponding files are then loaded from the `share/results/grids` folder.

    Note that the ROFF format is recommended for best performance.


    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        ensemble: str,
        roff_grid_name: str = None,
        roff_attribute_filter: List[str] = None,
        eclipse_grid_name: str = None,
        eclipse_init_parameters: List[str] = None,
        eclipse_restart_parameters: List[str] = None,
        initial_ijk_filter: Dict[str, int] = None,
    ):

        super().__init__(stretch=True)

        self.ensemble = webviz_settings.shared_settings["scratch_ensembles"][ensemble]

        if roff_grid_name:
            self.add_roff_grid_provider(
                grid_name=roff_grid_name, attribute_filter=roff_attribute_filter
            )
        elif (
            eclipse_grid_name and eclipse_init_parameters and eclipse_restart_parameters
        ):
            self.add_eclipse_grid_provider(
                grid_name=eclipse_grid_name,
                init_properties=eclipse_init_parameters,
                restart_properties=eclipse_restart_parameters,
            )
        else:
            raise ValueError(
                "Either provide roff_grid_name or "
                "eclipse_grid_name, eclipse_init_parameters and eclipse_restart_parameters"
            )
        self.add_store(
            ElementIds.IJK_CROP_STORE,
            storage_type=WebvizPluginABC.StorageType.SESSION,
        )

        self.add_view(
            View3D(
                grid_provider=self.grid_provider,
                grid_viz_service=self.grid_viz_service,
                initial_grid_filter=initial_ijk_filter if initial_ijk_filter else {},
            ),
            ElementIds.ID,
        )
        try:
            set_routes(app, self.grid_viz_service)
        except AssertionError:
            pass

    def add_roff_grid_provider(
        self, grid_name: str, attribute_filter: List[str] = None
    ) -> None:
        factory = EnsembleGridProviderFactory.instance()

        self.grid_provider: EnsembleGridProvider = factory.create_from_roff_files(
            ens_path=self.ensemble,
            grid_name=grid_name,
            attribute_filter=attribute_filter,
        )
        self.grid_viz_service = GridVizService.instance()
        self.grid_viz_service.register_provider(self.grid_provider)

    def add_eclipse_grid_provider(
        self, grid_name: str, init_properties: List[str], restart_properties: List[str]
    ) -> None:
        factory = EnsembleGridProviderFactory.instance()
        self.grid_provider = factory.create_from_eclipse_files(
            ens_path=self.ensemble,
            grid_name=grid_name,
            init_properties=init_properties,
            restart_properties=restart_properties,
        )
        self.grid_viz_service = GridVizService.instance()
        self.grid_viz_service.register_provider(self.grid_provider)



import warnings
from pathlib import Path
from typing import Dict, List

import pandas as pd
from fmu.tools.fipmapper import fipmapper


class VolumeValidatorAndCombinator:
    """
    Class to validate volumetric dataframes based on their type (static/dynamic),
    and to ensure correct format of combined volumetric dataframe before initialization
    of an InplaceVolumesModel instance. A best guess of each source's volume are made based
    on available columns. Only common columns between the different volumetric sources are kept.

    If the input contains both dynamic and static sources and a fipfile with FIPNUM to
    REGION∕ZONE mapping information is provided, the FipMapper from fmu.tools are utilized to
    create sets that are comparable in volumes. The volumes in the reulting dataframe will be
    combined per set. Only region selectors FIPNUM/REGION∕ZONE which are unique whithin each set
    are kept in the resulting dataframe, if none meets the criteria SET is used.

    Static dataframes must follow a strict format to fit the data processing steps of the
    InplaceVolumesModel instance. If custom columns found the source is set as dynamic to
    avoid this processing.
    Dynamic dataframes must contain a FIPNUM column but custom columns are allowed.
    """

    ENSEMBLE_COLUMNS = ["ENSEMBLE", "REAL", "SOURCE"]
    VALID_STATIC_SELECTORS = ["ZONE", "REGION", "FACIES", "LICENSE"]
    VALID_STATIC_RESPONSES = [
        "BULK",
        "NET",
        "PORV",
        "HCPV",
        "GIIP",
        "STOIIP",
        "ASSOCIATEDOIL",
        "ASSOCIATEDGAS",
    ]
    FLUID_TYPES = ["TOTAL", "GAS", "OIL"]

    def __init__(self, volumes_table: pd.DataFrame, fipfile: Path = None):

        self.volume_sources: Dict[str, List[str]] = {
            "static": [],
            "dynamic": [],
            "unknown": [],
        }
        self.disjoint_set_df = (
            fipmapper.FipMapper(yamlfile=fipfile).disjoint_sets() if fipfile else None
        )

        self.dframe = self.validate_and_combine_sources(
            self.drop_rows_with_totals_from_selectors(volumes_table)
        )
        self.volume_type = self.set_volumetric_type()

        if self.volume_type == "mixed":
            self.drop_total_columns()

    @property
    def possible_static_columns(self) -> list:
        possible_static_columns = self.VALID_STATIC_SELECTORS + self.ENSEMBLE_COLUMNS
        for fluid in self.FLUID_TYPES:
            possible_static_columns.extend(
                [f"{col}_{fluid}" for col in self.VALID_STATIC_RESPONSES]
            )
        return possible_static_columns

    def pore_to_porv_mapping(self, dframe: pd.DataFrame) -> pd.DataFrame:
        """
        Check if any columns startswith "PORE", if found rename to standard "PORV".
        Needs to be done on ensemble level to avoid issue with duplicate columns.
        """
        if any(col.startswith("PORE_") for col in dframe):
            dfs = []
            for _, df in dframe.groupby("ENSEMBLE"):
                dfs.append(
                    df.dropna(axis=1, how="all").rename(
                        columns={
                            f"PORE_{fluid}": f"PORV_{fluid}"
                            for fluid in self.FLUID_TYPES
                        },
                        errors="ignore",
                    )
                )
            dframe = pd.concat(dfs)
        return dframe

    def validate_and_combine_sources(self, volumes_table: pd.DataFrame) -> pd.DataFrame:
        """
        Validate columns for each volumetric source and combine them. Only common columns
        are kept. If a fipfile is provided volumes are summed per disjoint_set.
        """
        dfs = []
        all_columns = set()
        for source, voldf in volumes_table.groupby("SOURCE"):
            voldf = voldf.dropna(axis=1, how="all")
            voldf = self.pore_to_porv_mapping(voldf)
            volume_type = self.find_volume_type(voldf.columns, source)
            self.volume_sources[volume_type].append(source)
            all_columns.update(voldf.columns)
            dfs.append(voldf)

        dframe = (
            pd.concat(dfs, join="inner", ignore_index=True)
            if self.disjoint_set_df is None
            else self.create_set_dframe(volume_dfs=dfs)
        )

        missing_columns = [col for col in all_columns if col not in dframe]
        if missing_columns:
            warnings.warn(
                f"Skipping volumetric columns: {missing_columns} as they are not present in all "
                "volumetric sources"
            )
        return dframe

    def set_volumetric_type(self) -> str:
        """Return volumetric type based on available volume sources"""
        if self.volume_sources["static"] and (
            self.volume_sources["dynamic"] or self.volume_sources["unknown"]
        ):
            if self.volume_sources["dynamic"] and self.disjoint_set_df is None:
                raise ValueError(
                    "A fipfile must be provided to map FIPNUMS to REGION/ZONE when both static "
                    "and dynamic volume sources are given as input. "
                    f"Static sources: {self.volume_sources['static']} "
                    f"Dynamic sources: {self.volume_sources['dynamic']}"
                )
            return "mixed"
        return "static" if self.volume_sources["static"] else "dynamic"

    def find_volume_type(self, columns: list, source: str) -> str:
        """Return volume type (stativ/dynamic) based on columns and validate"""
        static_selectors_present = any(
            col in self.VALID_STATIC_SELECTORS for col in columns
        )
        non_static_columns = [
            col for col in columns if col not in self.possible_static_columns
        ]
        # Guess volume type based on columns
        if "FIPNUM" in columns:
            return "dynamic"

        if not non_static_columns:
            if not static_selectors_present:
                raise ValueError(
                    f"Static volume source {source} provided with no valid selectors. "
                    f"One of {self.VALID_STATIC_SELECTORS} must be provided."
                )
            return "static"

        if static_selectors_present:
            warnings.warn(
                f"The volumetric source {source} is considered a dynamic source due "
                f"to the presence of columns: {non_static_columns}. "
                "If this is a static source remove or rename these columns to "
                "trigger correct plugin mode. "
                f"Valid static columns are: {self.possible_static_columns}"
            )
            return "unknown"

        raise ValueError(
            f"Volume source {source} provided with no valid selectors. "
            f"One of {self.VALID_STATIC_SELECTORS + ['FIPNUM']} must be provided."
        )

    def create_set_dframe(self, volume_dfs: List[pd.DataFrame]) -> pd.DataFrame:
        """Sum Eclipse and RMS volumetrics over the common disjoints sets."""
        region_selectors = self.find_region_selectors()
        set_data_list = []
        for set_idx, df in self.disjoint_set_df.groupby("SET"):
            for voldf in volume_dfs:
                source = voldf["SOURCE"].unique()[0]
                if "FIPNUM" in voldf.columns:
                    filtered_df = voldf[voldf["FIPNUM"].isin(df["FIPNUM"].unique())]
                elif "ZONE" in voldf.columns and "REGION" in voldf.columns:
                    filtered_df = voldf[
                        (voldf["REGION"].isin(df["REGION"].unique()))
                        & (voldf["ZONE"].isin(df["ZONE"].unique()))
                    ]
                else:
                    raise ValueError(
                        f"Fipfile is provided but volumetric source {source} is missing "
                        "ZONE/REGION or FIPNUM definition."
                    )
                set_df = (
                    filtered_df.groupby(["ENSEMBLE", "REAL"])
                    .sum(numeric_only=True)
                    .reset_index()
                    .drop(labels=["FIPNUM", "REGION", "ZONE"], errors="ignore")
                )
                for col in region_selectors:
                    set_df[col] = df[col].iloc[0] if col != "SET" else set_idx
                set_df["SOURCE"] = source
                set_data_list.append(set_df)

        dframe = pd.concat(set_data_list, join="inner", ignore_index=True)
        if "FIPNUM" in dframe:
            dframe = dframe.sort_values(by=["FIPNUM"])
        return dframe

    def find_region_selectors(self) -> list:
        """Return region selectors that has a unique value
        per set. If none is found SET is used"""
        df = self.disjoint_set_df.groupby(["SET"]).nunique()
        regcols = ["FIPNUM", "REGION", "ZONE"]
        if any((df[x] == 1).all() for x in regcols):
            return [x for x in regcols if (df[x] == 1).all()]
        return ["SET"]

    def drop_total_columns(self) -> None:
        """Drop columns with "TOTAL" if both static and dynamic volumes in input"""
        total_columns = [col for col in self.dframe if col.endswith("_TOTAL")]
        if total_columns:
            warnings.warn(
                f"Dropping columns {total_columns} to avoid misleading comparison between "
                "static and dynamic columns with TOTAL."
            )
            self.dframe.drop(columns=total_columns, inplace=True)

    @staticmethod
    def drop_rows_with_totals_from_selectors(dframe: pd.DataFrame) -> pd.DataFrame:
        """Drop rows containing total volumes ("Totals") if present"""
        selectors = [col for col in ["ZONE", "REGION", "FACIES"] if col in dframe]
        for sel in selectors:
            dframe = dframe.loc[dframe[sel] != "Totals"]
        return dframe



import pathlib
from typing import Dict, List, Optional

import pandas as pd
import xtgeo
from webviz_config.webviz_store import webvizstore


def make_undefined_surface() -> xtgeo.RegularSurface:
    surf = xtgeo.RegularSurface(
        ncol=1, nrow=1, xinc=1, yinc=1
    )  # 1's as input is required
    surf.values = 0
    return surf


def surface_from_zone_prop(
    surface_table: pd.DataFrame, zone: str, prop: str, ensemble: str, stype: str
) -> xtgeo.RegularSurface:

    df = surface_table[
        (surface_table["zone"] == zone)
        & (surface_table["prop"] == prop)
        & (surface_table["ensemble"] == ensemble)
        & (surface_table["statistic"] == stype)
    ]
    if df.empty or len(df["path"].unique()) > 1:
        return make_undefined_surface()
    path = get_path(pathlib.Path(df["path"].unique()[0]))
    return xtgeo.surface_from_file(path.resolve())


@webvizstore
def generate_surface_table(
    zones: List[str],
    properties: List[str],
    ensembles: List,
    surface_renaming: Dict,
    surface_folders: Optional[Dict],
) -> pd.DataFrame:
    surface_folders = {} if not surface_folders else surface_folders
    statistics = ["mean", "stddev", "min", "max", "p10", "p90"]
    surface_table = []
    for ensemble in ensembles:
        for zone in zones:
            for prop in properties:
                for statistic in statistics:
                    zone_in_file = surface_renaming.get(zone, zone)
                    prop_in_file = surface_renaming.get(prop, prop)
                    path = (
                        surface_folders.get(ensemble, pathlib.Path())
                        / statistic
                        / f"{zone_in_file}--{prop_in_file}.gri"
                    )
                    if path.exists():
                        surface_table.append(
                            {
                                "ensemble": ensemble,
                                "zone": zone,
                                "prop": prop,
                                "statistic": statistic,
                                "path": str(path),
                            }
                        )
    return (
        pd.DataFrame(surface_table)
        if surface_table
        else pd.DataFrame(columns=["ensemble", "zone", "prop", "statistic", "path"])
    )


@webvizstore
def get_path(path: pathlib.Path) -> pathlib.Path:
    return path



from ._ensembles import EnsemblesSettings
from ._filter_realization import FilterRealizationSettings
from ._group_by import GroupBySettings
from ._resampling_frequency import ResamplingFrequencySettings
from ._time_series import TimeSeriesSettings
from ._visualization import VisualizationSettings



from typing import Dict, List, Tuple

from webviz_subsurface._providers import EnsembleSummaryProvider
from webviz_subsurface._utils.ensemble_summary_provider_set import (
    EnsembleSummaryProviderSet,
)

from .._types import DeltaEnsemble


def create_delta_ensemble_name(delta_ensemble: DeltaEnsemble) -> str:
    """Create delta ensemble name from delta ensemble"""
    name_a = delta_ensemble["ensemble_a"]
    name_b = delta_ensemble["ensemble_b"]
    return f"({name_a})-({name_b})"


def create_delta_ensemble_names(delta_ensembles: List[DeltaEnsemble]) -> List[str]:
    """Create list of delta ensemble names form list of delta ensembles"""
    return [
        create_delta_ensemble_name(delta_ensemble) for delta_ensemble in delta_ensembles
    ]


def create_delta_ensemble_name_dict(
    delta_ensembles: List[DeltaEnsemble],
) -> Dict[str, DeltaEnsemble]:
    """Create dictionary with delta ensemble name as key and and corresponding delta ensemble
    as value, from list if delta ensembles"""
    return {
        create_delta_ensemble_name(delta_ensemble): delta_ensemble
        for delta_ensemble in delta_ensembles
    }


def is_delta_ensemble_providers_in_provider_set(
    delta_ensemble: DeltaEnsemble, provider_set: EnsembleSummaryProviderSet
) -> bool:
    """Check if the delta ensemble providers exist in provider set

    `Returns:`
    * True if name of ensemble A and ensemble B exist among provider set names,
    false otherwise
    """
    return (
        delta_ensemble["ensemble_a"] in provider_set.provider_names()
        and delta_ensemble["ensemble_b"] in provider_set.provider_names()
    )


def create_delta_ensemble_provider_pair(
    delta_ensemble: DeltaEnsemble, provider_set: EnsembleSummaryProviderSet
) -> Tuple[EnsembleSummaryProvider, EnsembleSummaryProvider]:
    """Create pair of providers representing a delta ensemble

    `Return:`
    * Return Tuple with provider for ensemble A and ensemble B in a delta ensemble
    retrieved from provider set. If one or more provider does not exist, exception
    is raised!
    """

    ensemble_a = delta_ensemble["ensemble_a"]
    ensemble_b = delta_ensemble["ensemble_b"]
    if not is_delta_ensemble_providers_in_provider_set(delta_ensemble, provider_set):
        provider_names = provider_set.provider_names()
        raise ValueError(
            f"Request delta ensemble with ensemble {ensemble_a}"
            f" and ensemble {ensemble_b}. Ensemble {ensemble_a} exists: "
            f"{ensemble_a in provider_names}, ensemble {ensemble_b} exists: "
            f"{ensemble_b in provider_names}."
        )
    return (provider_set.provider(ensemble_a), provider_set.provider(ensemble_b))



from pathlib import Path
from typing import Callable, List, Optional, Tuple

import webviz_core_components as wcc
from webviz_config import WebvizPluginABC, WebvizSettings

from ._business_logic import SwatinitQcDataModel
from ._callbacks import plugin_callbacks
from ._layout import plugin_main_layout


class SwatinitQC(WebvizPluginABC):
    """This plugin is used to visualize the output from [check_swatinit]\
(https://fmu-docs.equinor.com/docs/subscript/scripts/check_swatinit.html) which is a QC tool
for Water Initialization in Eclipse runs when the `SWATINIT` keyword has been used. It is used to
quantify how much the volume changes from `SWATINIT` to `SWAT` at time zero in the dynamical model,
and help understand why it changes.

---
* **`csvfile`:** Path to an csvfile from check_swatinit. The path should be relative to the runpath
if ensemble and realization is given as input, if not the path needs to be absolute.
* **`ensemble`:** Which ensemble in `shared_settings` to visualize.
* **`realization`:** Which realization to pick from the ensemble
* **`faultlines`**: A csv file containing faultpolygons to be visualized together with the map view.
Export format from [xtgeo.xyz.polygons.dataframe](
https://xtgeo.readthedocs.io/en/latest/apiref/xtgeo.xyz.polygons.html#xtgeo.xyz.polygons.Polygons.dataframe
) \
[(example file)](\
https://github.com/equinor/webviz-subsurface-testdata/blob/master/01_drogon_ahm/\
realization-0/iter-0/share/results/polygons/toptherys--gl_faultlines_extract_postprocess.csv).

---
The `csvfile` can be generated by running the [CHECK_SWATINIT](https://fmu-docs.equinor.com/\
docs/ert/reference/forward_models.html?highlight=swatinit#CHECK_SWATINIT) forward model in ERT,
or with the "check_swatinit" command line tool.

"""

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        csvfile: str = "share/results/tables/check_swatinit.csv",
        ensemble: Optional[str] = None,
        realization: Optional[int] = None,
        faultlines: Path = None,
    ) -> None:
        super().__init__()

        self._datamodel = SwatinitQcDataModel(
            webviz_settings=webviz_settings,
            csvfile=csvfile,
            ensemble=ensemble,
            realization=realization,
            faultlines=faultlines,
        )
        self.add_webvizstore()
        self.set_callbacks()

    @property
    def layout(self) -> wcc.Tabs:
        return plugin_main_layout(self.uuid, self._datamodel)

    def set_callbacks(self) -> None:
        plugin_callbacks(self.uuid, self._datamodel)

    def add_webvizstore(self) -> List[Tuple[Callable, List[dict]]]:
        return self._datamodel.webviz_store



import json
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple

from dash import Dash, html
from webviz_config import WebvizPluginABC, WebvizSettings

from webviz_subsurface._providers import (
    EnsembleFaultPolygonsProviderFactory,
    EnsembleSurfaceProviderFactory,
)
from webviz_subsurface._providers.ensemble_fault_polygons_provider.fault_polygons_server import (
    FaultPolygonsServer,
)
from webviz_subsurface._providers.ensemble_surface_provider.surface_array_server import (
    SurfaceArrayServer,
)
from webviz_subsurface._utils.webvizstore_functions import read_csv

from ._tmp_well_pick_provider import WellPickProvider
from .callbacks import plugin_callbacks
from .color_tables import default_color_tables
from .layout import main_layout


class MapViewerFMU(WebvizPluginABC):
    """Surface visualizer for FMU ensembles.
A dashboard to covisualize arbitrary surfaces generated by FMU.

---

* **`ensembles`:** Which ensembles in `shared_settings` to visualize.
* **`attributes`:** List of surface attributes to include, if not given
    all surface attributes will be included.
* **`rel_surface_folder`:** Folder with maps relative to runpath in each realization.
    Default value is 'share/results/maps'.
* **`well_pick_file`:** A csv file with well picks.  See data input.
* **`fault_polygon_attribute`:** Which set of fault polygons to use.
* **`map_surface_names_to_well_pick_names`:** Allows mapping of file surface names
    to the relevant well pick name
* **`map_surface_names_to_fault_polygons`:** Allows mapping of file surface names
    to the relevant fault polygon set name
* **`color_tables`:** Color tables for the map layers

---
The available maps are gathered from the `share/results/maps/` folder
for each realization. Subfolders are not supported.

Observed maps are gathered from the `share/observations/maps/` folder in the case folder.<br>
The filenames need to follow a fairly strict convention, as the filenames are used as metadata:
`horizon_name--attribute--date` (`--date` is optional).<br> The files should be on `irap binary`
format with the suffix `.gri`.

The date is of the form `YYYYMMDD` or `YYYYMMDD_YYYYMMDD`, the latter would be for a delta
surface between two dates.<br>
See [this folder]\
(https://github.com/equinor/webviz-subsurface-testdata/tree/master/01_drogon_ahm/\
realization-0/iter-0/share/results/maps) \
for examples of file naming conventions.

Fault polygons are gathered from the `share/results/polygons` folder for each realization.<br>
Same file naming convention as for surfaces must be followed and the suffix should be `.pol`,
representing XYZ format usable by xtgeo.<br>
See [this file](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
01_drogon_ahm/realization-0/iter-0/share/results/polygons/\
toptherys--gl_faultlines_extract_postprocess.pol) for an example.

Well picks are provided as a csv file with columns `X_UTME,Y_UTMN,Z_TVDSS,MD,WELL,HORIZON`.
See [wellpicks.csv](https://github.com/equinor/webviz-subsurface-testdata/tree/master/\
observed_data/drogon_well_picks/wellpicks.csv) for an example.<br>
Well picks can be exported from RMS using this script: [extract_well_picks_from_rms.py]\
(https://github.com/equinor/webviz-subsurface-testdata/tree/master/observed_data/\
drogon_well_picks/extract_well_picks_from_rms.py)

See     [this file]\
https://raw.githubusercontent.com/emerson-eps/color-tables/main/react-app/src/component/\
color-tables.json for color_tables format.

"""

    # pylint: disable=too-many-arguments, too-many-locals
    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        ensembles: list,
        attributes: list = None,
        well_pick_file: Path = None,
        fault_polygon_attribute: Optional[str] = None,
        map_surface_names_to_fault_polygons: Dict[str, str] = None,
        map_surface_names_to_well_pick_names: Dict[str, str] = None,
        rel_surface_folder: str = "share/results/maps",
        color_tables: Path = None,
        hillshading_enabled: bool = None,
    ):

        super().__init__()
        self._hillshading = hillshading_enabled
        surface_provider_factory = EnsembleSurfaceProviderFactory.instance()
        fault_polygons_provider_factory = (
            EnsembleFaultPolygonsProviderFactory.instance()
        )

        self._ensemble_surface_providers = {
            ens: surface_provider_factory.create_from_ensemble_surface_files(
                webviz_settings.shared_settings["scratch_ensembles"][ens],
                attribute_filter=attributes,
                rel_surface_folder=rel_surface_folder,
            )
            for ens in ensembles
        }
        self._surface_server = SurfaceArrayServer.instance(app)

        self.well_pick_provider = None
        self.well_pick_file = well_pick_file
        if self.well_pick_file is not None:
            well_pick_table = read_csv(self.well_pick_file)
            self.well_pick_provider = WellPickProvider(
                dframe=well_pick_table,
                map_surface_names_to_well_pick_names=map_surface_names_to_well_pick_names,
            )
            self.well_pick_provider.get_geojson(
                self.well_pick_provider.well_names(), "TopVolantis"
            )

        self._ensemble_fault_polygons_providers = {
            ens: fault_polygons_provider_factory.create_from_ensemble_fault_polygons_files(
                webviz_settings.shared_settings["scratch_ensembles"][ens]
            )
            for ens in ensembles
        }
        all_fault_polygon_attributes = self._ensemble_fault_polygons_providers[
            ensembles[0]
        ].attributes()
        self.fault_polygon_attribute: Optional[str] = None
        if (
            fault_polygon_attribute is not None
            and fault_polygon_attribute in all_fault_polygon_attributes
        ):
            self.fault_polygon_attribute = fault_polygon_attribute
        elif all_fault_polygon_attributes:
            self.fault_polygon_attribute = all_fault_polygon_attributes[0]
        else:
            self.fault_polygon_attribute = None

        self._fault_polygons_server = FaultPolygonsServer.instance(app)
        for fault_polygons_provider in self._ensemble_fault_polygons_providers.values():
            self._fault_polygons_server.add_provider(fault_polygons_provider)

        self.map_surface_names_to_fault_polygons = (
            map_surface_names_to_fault_polygons
            if map_surface_names_to_fault_polygons is not None
            else {}
        )
        self.color_tables = (
            default_color_tables
            if color_tables is None
            else json.loads(color_tables.read_text())
        )
        self.set_callbacks()

    @property
    def layout(self) -> html.Div:
        reals = []
        for provider in self._ensemble_surface_providers.values():
            reals.extend([x for x in provider.realizations() if x not in reals])
        return main_layout(
            get_uuid=self.uuid,
            well_names=self.well_pick_provider.well_names()
            if self.well_pick_provider is not None
            else [],
            realizations=reals,
            color_tables=self.color_tables,
        )

    def set_callbacks(self) -> None:

        plugin_callbacks(
            get_uuid=self.uuid,
            ensemble_surface_providers=self._ensemble_surface_providers,
            surface_server=self._surface_server,
            ensemble_fault_polygons_providers=self._ensemble_fault_polygons_providers,
            fault_polygon_attribute=self.fault_polygon_attribute,
            fault_polygons_server=self._fault_polygons_server,
            map_surface_names_to_fault_polygons=self.map_surface_names_to_fault_polygons,
            well_picks_provider=self.well_pick_provider,
            color_tables=self.color_tables,
            plugin_data_output=self.plugin_data_output,
            plugin_data_requested=self.plugin_data_requested,
        )

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        store_functions = []
        if self.well_pick_file is not None:
            store_functions.append((read_csv, [{"csv_file": self.well_pick_file}]))
        return store_functions



########################################
#
#  Copyright (C) 2021-     Equinor ASA
#
#  webviz-subsurface is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  webviz-subsurface is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http:#www.gnu.org/licenses/gpl.html>
#  for more details.
#
########################################
########################################
#
#  The code in this file is based on and/or inspired by opm-common,
#  which is distributed under the GNU General Public License v3.0,
#  and available at https://github.com/OPM/opm-common.
#
#  Especially in the following points this file, or files imported in this file
#  within the same context, make use of or are based on methods developed
#  for the opm-common project:
#  - Reading of Eclipse INIT files
#  - Units, unit systems and conversion of units
#  - Class structures for storing and accessing PVT data,
#    both base classes and derived classes for specific fluids
#
#  You should have received a copy of the GNU General Public License
#  along with webviz-subsurface. If not, see (https://www.gnu.org/licenses/gpl-3.0.html).
#
#  (12-01-2021)
#
########################################

from typing import Any, Callable, Optional, Tuple, Union

import numpy as np

# opm is only available for Linux,
# hence, ignore any import exception here to make
# it still possible to use the PvtPlugin on
# machines with other OSes.
#
# NOTE: Functions in this file cannot be used
#       on non-Linux OSes.
try:
    from opm.io.ecl import EclFile
except ImportError:
    pass

from ..eclipse_unit import ConvertUnits, CreateUnitConverter, EclUnitEnum, EclUnits
from .pvt_common import (
    EclPhaseIndex,
    EclPropertyTableRawData,
    FluidImplementation,
    InitFileDefinitions,
    PVDx,
    PVTx,
    PvxOBase,
    is_const_compr_index,
    surface_mass_density,
)


class LiveOil(PvxOBase):
    """Class holding a PVT interpolant and access methods for PVT data for live oil

    Attributes:
        interpolant: The interpolant object
    """

    def __init__(
        self,
        index_table: int,
        raw: EclPropertyTableRawData,
        surface_mass_density_oil: float,
        surface_mass_density_gas: float,
        convert: Tuple[
            Callable[
                [
                    float,
                ],
                float,
            ],
            ConvertUnits,
        ],
    ) -> None:
        """Initializes a LiveOil object.

        Creates an interpolant for live oil for the Eclipse data given
        by the raw data in the table with the given index.

        Args:
            index_table: Index of the PVT table
            raw: Eclipse raw data object
            surface_mass_density_oil: Surface mass density of oil
            surface_mass_density_gas: Surface mass density of gas
            convert: Tuple holding a callable and a ConvertUnits object for unit conversions

        """
        self.__interpolant = PVTx(index_table, raw, convert)
        self.__surface_mass_density_oil = surface_mass_density_oil
        self.__surface_mass_density_gas = surface_mass_density_gas

    def formation_volume_factor(
        self, ratio: np.ndarray, pressure: np.ndarray
    ) -> np.ndarray:
        """Args:
            ratio: Ratio (key) values the volume factor values are requested for.
            pressure: Pressure values the volume factor values are requested for.

        Returns:
            Volume factor values for the given ratio and pressure values.

        """
        return self.__interpolant.formation_volume_factor(ratio, pressure)

    def viscosity(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Args:
            ratio: Ratio (key) values the viscosity values are requested for.
            pressure: Pressure values the viscosity values are requested for.

        Returns:
            Viscosity values for the given ratio and pressure values.

        """
        return self.__interpolant.viscosity(ratio, pressure)

    def density(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Args:
            ratio: Ratio (key) values the density values are requested for.
            pressure: Pressure values the density values are requested for.

        Returns:
            Density values for the given ratio and pressure values.

        """
        # rho_o = (rho_o,sc + Rs * rho_g,sc) / B_o
        fvf_oil = self.formation_volume_factor(ratio, pressure)
        return np.array(
            [
                (
                    self.__surface_mass_density_oil
                    + ratio_sample * self.__surface_mass_density_gas
                )
                / fvf_oil_sample
                for ratio_sample, fvf_oil_sample in zip(ratio, fvf_oil)
            ]
        )

    def get_keys(self) -> np.ndarray:
        """Returns all primary key values (Rs)"""
        return self.__interpolant.get_keys()

    def get_independents(self) -> np.ndarray:
        """Returns all independent pressure values (Po)"""
        return self.__interpolant.get_independents()


class DeadOil(PvxOBase):
    """Class holding a PVT interpolant and access methods for PVT data for dead oil

    Attributes:
        interpolant: The interpolant object
    """

    def __init__(
        self,
        table_index: int,
        raw: EclPropertyTableRawData,
        surface_mass_density_oil: float,
        convert: ConvertUnits,
    ) -> None:
        """Initializes a DeadOil object.

        Creates an interpolant for dead oil for the Eclipse data given
        by the raw data in the table with the given index.

        Args:
            index_table: Index of the PVT table
            raw: Eclipse raw data object
            surface_mass_density_oil: Surface mass density of oil
            convert: ConvertUnits object for unit conversions

        """
        self.__interpolant = PVDx(table_index, raw, convert)
        self.__surface_mass_density_oil = surface_mass_density_oil

    def formation_volume_factor(
        self, ratio: np.ndarray, pressure: np.ndarray
    ) -> np.ndarray:
        """Args:
            ratio: Dummy argument, only to conform to interface of base class.
            pressure: Pressure values the volume factor values are requested for.

        Returns:
            Volume factor values for the given pressure values.

        """
        return self.__interpolant.formation_volume_factor(pressure)

    def viscosity(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Args:
            ratio: Dummy argument, only to conform to interface of base class.
            pressure: Pressure values the viscosity values are requested for.

        Returns:
            Viscosity values for the given pressure values.

        """
        return self.__interpolant.viscosity(pressure)

    def density(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Args:
            ratio: Dummy argument, only to conform to interface of base class.
            pressure: Pressure values the density values are requested for.

        Returns:
            Density values for the given ratio and pressure values.

        """
        # rho_o = rho_o,sc / B_o
        fvf_oil = self.formation_volume_factor(ratio, pressure)
        return np.array(
            [
                self.__surface_mass_density_oil / fvf_oil_sample
                for fvf_oil_sample in fvf_oil
            ]
        )

    def get_keys(self) -> np.ndarray:
        """Returns all primary keys.

        Since this is dead oil, there is no dependency on Rs.
        Hence, this method returns a list holding floats of value 0.0
        for each independent value.

        """
        return self.__interpolant.get_keys()

    def get_independents(self) -> np.ndarray:
        """Returns all independent pressure values (Po)"""
        return self.__interpolant.get_independents()


class DeadOilConstCompr(PvxOBase):
    """Class holding a PVT interpolant and access methods for PVT data
    for dead oil with constant compressibility"""

    def __init__(
        self,
        index_table: int,
        raw: EclPropertyTableRawData,
        convert: ConvertUnits,
        surface_mass_density_oil: float,
    ) -> None:
        """Initializes a DeadOilConstCompr object.

        Creates an interpolant for dead oil with constant compressibility for the Eclipse data given
        by the raw data in the table with the given index.

        Args:
            index_table: Index of the PVT table
            raw: Eclipse raw data
            convert: Tuple holding a callable and a ConvertUnits object for unit conversions

        Raises:
            ValueError in case of an invalid reference oil pressure
        """
        self.__surface_mass_density_oil = surface_mass_density_oil

        column_stride = raw.num_rows * raw.num_tables * raw.num_primary
        table_stride = raw.num_primary * raw.num_rows
        current_stride = index_table * table_stride

        self.__fvf_ref = convert.column[0](
            raw.data[0 * column_stride + current_stride]
        )  # Bo
        self.__c_o_ref = convert.column[1](
            raw.data[1 * column_stride + current_stride]
        )  # Co
        self.__visc_ref = convert.column[2](
            raw.data[2 * column_stride + current_stride]
        )  # mu_o
        self.__c_v_ref = convert.column[3](
            raw.data[3 * column_stride + current_stride]
        )  # Cv

        self.__p_o_ref = convert.independent(raw.data[current_stride])

        if abs(self.__p_o_ref) < 1.0e20:
            raise ValueError("Invalid Input PVCDO Table")

    def formation_volume_factor(
        self, ratio: np.ndarray, pressure: np.ndarray
    ) -> np.ndarray:
        """Computes formation volume factor values
        for the given pressure values.

        Args:
            ratio: Dummy argument, only to conform to interface of base class
            pressure: Pressure values

        Returns:
            Formation volume factor values for the given pressure values.

        """
        return self.__evaluate(pressure, lambda p: 1.0 / self.__recip_fvf(p))

    def viscosity(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Computes viscosity values for the given pressure values.

        Args:
            ratio: Dummy argument, only to conform to interface of base class
            pressure:Pressure values

        Returns:
            Viscosity values corresponding
            to the given pressure values.

        """
        return self.__evaluate(
            pressure, lambda p: self.__recip_fvf(p) / self.__recip_fvf_visc(p)
        )

    def __recip_fvf(self, p_o: float) -> float:
        """Computes the reciprocal of the formation volume factor for the given oil pressure.

        Args:
            p_o: Oil pressure

        Returns:
            Reciprocal of the formation volume factor

        """

        # B_o(P) = B_o(P_ref) * e^-x
        # x = C * (P - P_ref)
        # 1 / B_o(P) = 1 / B_o(P_ref) * e^x

        x = self.__c_o_ref * (p_o - self.__p_o_ref)

        return self.__exp(x) / self.__fvf_ref

    def density(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Args:
            ratio: Ratio (key) values the density values are requested for.
            pressure: Pressure values the density values are requested for.

        Returns:
            Density values for the given ratio and pressure values.

        """
        # rho_o = rho_o,sc / B_o
        fvf_oil = self.formation_volume_factor(ratio, pressure)
        return np.array(
            [
                self.__surface_mass_density_oil / fvf_oil_sample
                for fvf_oil_sample in fvf_oil
            ]
        )

    def __recip_fvf_visc(self, p_o: float) -> float:
        """Computes the reciprocal of the product of formation volume factor
        and viscosity for the given oil pressure.

        Args:
            p_o: Oil pressure

        Returns:
            Reciprocal of the product of formation volume factor and viscosity

        """

        # mu_o(P) * B_o(P) = B_o(P_ref) * mu_o(P_ref) * e^-y
        # y = (C - C_v) * (P - P_ref)
        # 1 / (mu_o(P) * B_o(P)) = 1 / (B_o(P_ref) * mu_o(P_ref)) * e^y
        y = (self.__c_o_ref - self.__c_v_ref) * (p_o - self.__p_o_ref)

        return self.__exp(y) / (self.__fvf_ref * self.__visc_ref)

    @staticmethod
    def __exp(x: float) -> float:
        """Internal helper function.

        Computes an approximation to the exponential function by means
        of a Taylor series (to second degree).

        Args:
            x: Expression to use (see manual)

        Returns:
            Approximation to exponential function

        """
        return 1.0 + x * (1.0 + x / 2.0)

    @staticmethod
    def __evaluate(
        pressures: np.ndarray, calculate: Callable[[Any], Any]
    ) -> np.ndarray:
        """Calls the calculate method with each of pressure values and returns the results.

        Args:
            pressures: Pressure values
            calculate: Method to be called with each of the pressure values

        Returns:
            Result values

        """
        return np.array([calculate(pressure) for pressure in pressures])

    def get_keys(self) -> np.ndarray:
        """Returns all primary keys.

        Since this is dead oil, there is no dependency on any ratio.
        Hence, this method returns a list holding a single float of value 0.0.

        """
        return np.zeros(1)

    def get_independents(self) -> np.ndarray:
        """Returns all independent pressure values (Po).

        Since this is water, this does return with only one single pressure value,
        the reference pressure.
        """
        return np.array(
            [
                self.__p_o_ref,
            ]
        )


class Oil(FluidImplementation):
    """Class for storing PVT tables for oil

    Holds a list of regions (one per PVT table).

    Attributes:
        surface_mass_densities: Surface mass densities
        keep_unit_system: True if the original unit system was kept
        original_unit_system: An ErtEclUnitEnum representing the original unit system
    """

    def __init__(
        self,
        raw: EclPropertyTableRawData,
        unit_system: int,
        is_const_compr: bool,
        surface_mass_densities: Tuple[np.ndarray, np.ndarray],
        keep_unit_system: bool = True,
    ):
        """Initializes an Oil object.

        Args:
            raw: Eclipse raw data object
            unit_system: The original unit system
            is_const_compr: True if oil has constant compressibility, else False
            surface_mass_densities: Surface mass densities
            keep_unit_system:
                True if the original unit system shall be kept,
                False if units shall be converted to SI units.

        """
        super().__init__(unit_system, keep_unit_system)
        self.surface_mass_densities = surface_mass_densities
        self.original_unit_system = EclUnitEnum(unit_system)
        self.create_pvt_interpolants(raw, is_const_compr, unit_system)

    def formation_volume_factor_unit(self, latex: bool = False) -> str:
        """Creates and returns the unit symbol of the formation volume factor.

        Args:
            latex: True if the unit symbol shall be returned as LaTeX, False if not.

        Returns:
            The unit symbol of the formation volume factor.

        """
        unit_system = EclUnits.create_unit_system(
            self.original_unit_system
            if self.keep_unit_system
            else EclUnitEnum.ECL_SI_UNITS
        )

        if latex:
            return (
                rf"${{{unit_system.reservoir_volume().symbol}}}/"
                rf"{{{unit_system.surface_volume_liquid().symbol}}}$"
            )
        return (
            f"{unit_system.reservoir_volume().symbol}/"
            f"{unit_system.surface_volume_liquid().symbol}"
        )

    def ratio_unit(self, latex: bool = False) -> str:
        """Creates and returns the unit symbol of the phase ratio.

        Args:
            latex: True if the unit symbol shall be returned as LaTeX, False if not.

        Returns:
            The unit symbol of the phase ratio.

        """
        unit_system = EclUnits.create_unit_system(
            self.original_unit_system
            if self.keep_unit_system
            else EclUnitEnum.ECL_SI_UNITS
        )

        if latex:
            return (
                rf"${{{unit_system.surface_volume_gas().symbol}}}/"
                rf"{{{unit_system.surface_volume_liquid().symbol}}}$"
            )
        return (
            f"{unit_system.surface_volume_gas().symbol}/"
            f"{unit_system.surface_volume_liquid().symbol}"
        )

    def dead_oil_unit_converter(
        self, unit_system: Union[int, EclUnits.UnitSystem]
    ) -> ConvertUnits:
        """Creates a ConvertUnits object for unit conversions for dead oil.

        Args:
            unit_system:
                Either an integer or an enum
                describing the unit system the units are stored in

        Returns:
            ConvertUnits object for unit conversions.

        """
        if not isinstance(unit_system, EclUnits.UnitSystem):
            unit_system = EclUnits.create_unit_system(unit_system)

        return super().pvdx_unit_converter() or ConvertUnits(
            CreateUnitConverter.ToSI.pressure(unit_system),
            [
                CreateUnitConverter.ToSI.recip_fvf(unit_system),
                CreateUnitConverter.ToSI.recip_fvf_visc(unit_system),
                CreateUnitConverter.ToSI.recip_fvf_deriv_press(unit_system),
                CreateUnitConverter.ToSI.recip_fvf_visc_deriv_press(unit_system),
            ],
        )

    def pvcdo_unit_converter(
        self, unit_system: Union[int, EclUnits.UnitSystem]
    ) -> ConvertUnits:
        """Creates a ConvertUnits object for unit conversions
        for dead oil with constant compressibility.

        Args:
            unit_system:
                Either an integer or an enum
                describing the unit system the units are stored in

        Returns:
            ConvertUnits object for unit conversions.

        """
        if not isinstance(unit_system, EclUnits.UnitSystem):
            unit_system = EclUnits.create_unit_system(unit_system)

        return super().pvdx_unit_converter() or ConvertUnits(
            CreateUnitConverter.ToSI.pressure(unit_system),
            [
                CreateUnitConverter.ToSI.fvf(unit_system),
                CreateUnitConverter.ToSI.compressibility(unit_system),
                CreateUnitConverter.ToSI.viscosity(unit_system),
                CreateUnitConverter.ToSI.compressibility(unit_system),
            ],
        )

    def live_oil_unit_converter(
        self, unit_system: Union[int, EclUnits.UnitSystem]
    ) -> Tuple[Callable[[float,], float,], ConvertUnits]:
        """Creates a tuple consisting of a callable and a ConvertUnits object
        for unit conversions for live oil.

        Args:
            unit_system:
                Either an integer or an enum
                describing the unit system the units are stored in

        Returns:
            Tuple consisting of callable and ConvertUnits object

        """
        if not isinstance(unit_system, EclUnits.UnitSystem):
            unit_system = EclUnits.create_unit_system(unit_system)

        return super().pvtx_unit_converter() or (
            CreateUnitConverter.ToSI.dissolved_gas_oil_ratio(unit_system),
            self.dead_oil_unit_converter(unit_system),
        )

    def create_pvt_interpolants(
        self, raw: EclPropertyTableRawData, is_const_compr: bool, unit_system: int
    ) -> None:
        """Creates interpolants for PVT data based on the type of the oil
        (i.e. live, dead and dead with constant compressibility).

        Args:
            raw: Eclipse raw data object
            unit_system: Number describing the unit system the values in the raw data are stored in

        Raises:
            A FluidImplementation.InvalidArgument exception
            in case of a size mismatch in the raw data.

        """
        if raw.num_primary == 0:
            raise super().InvalidArgument("Oil PVT table without primary lookup key")
        if raw.num_cols != 5:
            raise super().InvalidArgument("PVT table for oil must have five columns")
        if len(raw.primary_key) != (raw.num_primary * raw.num_tables):
            raise super().InvalidArgument(
                "Size mismatch in RS nodes of PVT table for oil"
            )
        if len(raw.data) != (
            raw.num_primary * raw.num_rows * raw.num_cols * raw.num_tables
        ):
            raise super().InvalidArgument(
                "Size mismatch in Condensed table data of PVT table for oil"
            )

        if raw.num_primary == 1:
            self.create_dead_oil(raw, is_const_compr, unit_system)
        else:
            self.create_live_oil(raw, unit_system)

    def create_live_oil(self, raw: EclPropertyTableRawData, unit_system: int) -> None:
        """Creates interpolants for live oil from the given raw Eclipse data and uses
        a live oil unit converter based on the given unit system.

        Args:
            raw: Eclipse raw data object
            unit_system: Integer representation of the unit system used in Eclipse data

        """
        cvrt = self.live_oil_unit_converter(unit_system)

        self._regions = self.make_interpolants_from_raw_data(
            # PKey   Inner   C0     C1         C2           C3
            # Rs     Po      1/B    1/(B*mu)   d(1/B)/dPo   d(1/(B*mu))/dPo
            #        :       :      :          :            :
            raw,
            lambda table_index, raw: LiveOil(
                table_index,
                raw,
                self.surface_mass_densities[0][table_index],
                self.surface_mass_densities[1][table_index],
                cvrt,
            ),
        )

        if len(self.surface_mass_densities[0]) != len(self._regions) or len(
            self.surface_mass_densities[1]
        ) != len(self._regions):
            raise ValueError(
                (
                    "The given Eclipse INIT file seems to be broken"
                    "(number of surface density values does not equal number of PVT regions)."
                )
            )

    def create_dead_oil(
        self, raw: EclPropertyTableRawData, const_compr: bool, unit_system: int
    ) -> None:
        """Creates interpolants for dead oil (with and without constant compressibility)
        from the given raw Eclipse data and uses a dead oil (or pvcdo) unit converter
        based on the given unit system.

        Args:
            raw: Eclipse raw data object
            unit_system: Integer representation of the unit system used in Eclipse data

        """
        if const_compr:
            cvrt = self.pvcdo_unit_converter(unit_system)

            self._regions = self.make_interpolants_from_raw_data(
                raw,
                lambda table_index, raw: DeadOilConstCompr(
                    table_index, raw, cvrt, self.surface_mass_densities[0][table_index]
                ),
            )
            return

        cvrt = self.dead_oil_unit_converter(unit_system)

        self._regions = self.make_interpolants_from_raw_data(
            raw,
            lambda table_index, raw: DeadOil(
                table_index, raw, self.surface_mass_densities[0][table_index], cvrt
            ),
        )

    def is_live_oil(self) -> bool:
        """Checks if this fluid is live oil.

        Returns:
            True if live oil, else False

        """
        if len(self._regions) > 0:
            return isinstance(self._regions[0], LiveOil)
        return False

    def is_dead_oil(self) -> bool:
        """Checks if this fluid is dead oil with variable compressibility.

        Returns:
            True if dead oil with variable compressibility, else False

        """
        if len(self._regions) > 0:
            return isinstance(self._regions[0], DeadOil)
        return False

    def is_dead_oil_const_compr(self) -> bool:
        """Checks if this fluid is dead oil with constant compressibility.

        Returns:
            True if dead oil with constant compressibility, else False

        """
        if len(self._regions) > 0:
            return isinstance(self._regions[0], DeadOilConstCompr)
        return False

    @staticmethod
    def from_ecl_init_file(
        ecl_init_file: EclFile, keep_unit_system: bool = False
    ) -> Optional["Oil"]:
        """Reads the given Eclipse file and creates an Oil object from its data.

        Args:
            ecl_init_file: Eclipse INIT file
            keep_unit_system:
                Set to True if the unit system used in the Eclipse file
                shall be kept, False if SI shall be used.

        Returns:
            An Oil object or None if the data in the Eclipse file was invalid

        """
        intehead = ecl_init_file.__getitem__(InitFileDefinitions.INTEHEAD_KW)

        logihead = ecl_init_file.__getitem__(InitFileDefinitions.LOGIHEAD_KW)
        is_is_const_compr = bool(logihead[is_const_compr_index()])

        raw = EclPropertyTableRawData()

        tab_dims = ecl_init_file.__getitem__("TABDIMS")
        tab = ecl_init_file.__getitem__("TAB")

        num_rs = tab_dims[InitFileDefinitions.TABDIMS_NRPVTO_ITEM]

        raw.num_rows = tab_dims[InitFileDefinitions.TABDIMS_NPPVTO_ITEM]
        raw.num_cols = 5
        raw.num_tables = tab_dims[InitFileDefinitions.TABDIMS_NTPVTO_ITEM]

        if raw.num_tables == 0:
            return None

        if logihead[InitFileDefinitions.LOGIHEAD_RS_INDEX]:
            raw.num_primary = num_rs

        else:
            raw.num_primary = 1

        num_tab_elements = raw.num_primary * raw.num_tables
        start = tab_dims[InitFileDefinitions.TABDIMS_JBPVTO_OFFSET_ITEM] - 1
        raw.primary_key = tab[start : start + num_tab_elements]

        num_tab_elements = (
            raw.num_primary * raw.num_rows * raw.num_cols * raw.num_tables
        )
        start = tab_dims[InitFileDefinitions.TABDIMS_IBPVTO_OFFSET_ITEM] - 1
        raw.data = tab[start : start + num_tab_elements]

        surface_mass_densities = (
            surface_mass_density(ecl_init_file, EclPhaseIndex.LIQUID, keep_unit_system),
            surface_mass_density(ecl_init_file, EclPhaseIndex.VAPOUR, keep_unit_system),
        )

        return Oil(
            raw,
            intehead[InitFileDefinitions.INTEHEAD_UNIT_INDEX],
            is_is_const_compr,
            surface_mass_densities,
            keep_unit_system,
        )



from webviz_config.utils import StrEnum


class PressurePlotMode(StrEnum):
    MEAN = "mean"
    SINGLE_REAL = "single-real"


class NodeType(StrEnum):
    WELL = "well"
    GROUP = "group"
    WELL_BH = "well-bh"


class ChartType(StrEnum):
    BAR = "bar"
    PIE = "pie"
    AREA = "area"



import json
import warnings
from pathlib import Path
from typing import Callable, List, Tuple
from uuid import uuid4

import pandas as pd
from dash import html
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore
from webviz_subsurface_components import Map

from .._datainput.fmu_input import scratch_ensemble


class SubsurfaceMap(WebvizPluginABC):
    """Visualizes reservoir grids in a map view, additionally it can
visualize the flow pattern in the simulation output using streamlines.
Input can be either a premade json object or data can be extracted from
a FMU ensemble.

---
**Two input options: Ensemble data or premade json file**

**json file**
* **`jsonfile`:** jsonfile with data, suitable for the \
corresponding [subsurface map component]\
(https://github.com/equinor/webviz-subsurface-components) \
 (absolute path or relative to config file).

**Ensemble data**
* **`ensemble`:** Which ensemble in `shared_settings` to visualize (**just one**).
* **`map_value`:** Which property to show in the map (e.g. `PERMX`).
* **`flow_value`:** Which property to use for the streamlines animation
  (e.g. `FLOWAT`).
* **`time_step`:** Which report or time step to use in the simulation output.

---

For ensemble data input, the key `FLORES` needs to be in the `RPTRST` keyword of the simulation
data deck for flow fields like `FLOWAT` and `FLOOIL` to be included in the data.

?> Using the ensemble method, the cell-by-cell mean values over all the grids in the ensemble are \
used, both for properties and flow fields. A consequence of this is that all the grids in the \
ensemble have to be equal (though the properties can vary), meaning that e.g. structural \
uncertainty unfortunately is not supported. Taking the cell-by-cell will also tend to \
give less property variations than you would see in a single realization. \
To look at a single realization you currently have to define a separate ensemble \
consisting of just a single realization.

!> Using the ensemble method, `UNRST` and `INIT` files are autodetected in the realizations under \
`eclipse/model`. You should therefore not have more than one of each of these files to make sure \
that you are reading the correct data.
"""

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        jsonfile: Path = None,
        ensemble: str = None,
        map_value: str = None,
        flow_value: str = None,
        time_step: int = None,
    ):
        # TODO(Sigurd) What is the correct type(s) for time_step?

        super().__init__()

        self.jsonfile = jsonfile if jsonfile else None

        if jsonfile and ensemble:
            raise ValueError(
                'Incorrent arguments. Either provide a "jsonfile" or "ensemble", "map_value" '
                '"flow_value" and "time_step"'
            )
        if jsonfile:
            with open(get_path(jsonfile), "r") as filehandle:
                self.map_data = json.dumps(json.load(filehandle))
        elif ensemble and map_value and flow_value and time_step:
            self.map_value = map_value
            self.flow_value = flow_value
            self.time_step = time_step
            self.ensemble_path = webviz_settings.shared_settings["scratch_ensembles"][
                ensemble
            ]
            self.map_data = get_map_data(
                self.ensemble_path, self.map_value, self.flow_value, self.time_step
            )
        else:
            raise ValueError(
                'Incorrent arguments. Either provide a "jsonfile" or "ensemble", "map_value" '
                '"flow_value" and "time_step"'
            )
        self.map_id = f"map-{uuid4()}"

    @property
    def layout(self) -> html.Div:
        return html.Div([Map(id=self.map_id, data=self.map_data)])

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return [
            (get_path, [{"path": Path(self.jsonfile)}])
            if self.jsonfile
            else (
                get_uncompressed_data,
                [
                    {
                        "ensemble_path": self.ensemble_path,
                        "map_value": self.map_value,
                        "flow_value": self.flow_value,
                        "time_step": self.time_step,
                    }
                ],
            )
        ]


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_map_data(
    ensemble_path: str, map_value: str, flow_value: str, time_step: int
) -> str:
    """Returns map data in the format of a JSON string, suitable for the
    corresponding subsurface map component in
    https://github.com/equinor/webviz-subsurface-components
    """
    # pylint: disable=too-many-locals

    grid = get_uncompressed_data(ensemble_path, map_value, flow_value, time_step)

    indices_col = ["i", "j", "k"]
    x_col = ["x0", "x1", "x2", "x3"]
    y_col = ["y0", "y1", "y2", "y3"]
    flow_col = ["FLOWI+", "FLOWJ+"]

    resolution = 1000

    grid = grid[indices_col + x_col + y_col + ["value"] + flow_col]
    grid = grid[grid["value"] > 0]

    xmin, xmax = grid[x_col].values.min(), grid[x_col].values.max()
    ymin, ymax = grid[y_col].values.min(), grid[y_col].values.max()

    flowmin, flowmax = grid[flow_col].values.min(), grid[flow_col].values.max()

    valmin, valmax = grid["value"].min(), grid["value"].max()

    if (xmax - xmin) > (ymax - ymin):
        coord_scale = resolution / (xmax - xmin)
    else:
        coord_scale = resolution / (ymax - ymin)

    grid[x_col] = (grid[x_col] - xmin) * coord_scale
    grid[y_col] = (grid[y_col] - ymin) * coord_scale
    grid[x_col + y_col] = grid[x_col + y_col].astype(int)

    flow_scale = resolution / (flowmax - flowmin)
    grid[flow_col] = (grid[flow_col] - flowmin) * flow_scale
    grid[flow_col] = grid[flow_col].astype(int)

    val_scale = resolution / (valmax - valmin)
    grid["value"] = (grid["value"] - valmin) * val_scale
    grid["value"] = grid["value"].astype(int)

    grid[indices_col] = grid[indices_col].astype(int)

    data = {
        "values": grid.values.tolist(),
        "linearscales": {
            "coord": [float(coord_scale), float(xmin), float(ymin)],
            "value": [float(val_scale), float(valmin)],
            "flow": [float(flow_scale), float(flowmin)],
        },
    }

    return json.dumps(data, separators=(",", ":"))


@webvizstore
def get_uncompressed_data(
    ensemble_path: str, map_value: str, flow_value: str, time_step: int
) -> pd.DataFrame:

    ens = scratch_ensemble("", ensemble_path)

    properties = [map_value, f"{flow_value}I+", f"{flow_value}J+"]
    if "PERMX" not in properties:
        properties.append("PERMX")

    with warnings.catch_warnings():
        # Grid support outside initial scope of fmu-ensemble v2. Need to consider options
        warnings.filterwarnings("ignore", category=FutureWarning, module="fmu.ensemble")
        grid = ens.get_eclgrid(properties, report=time_step)

    grid = grid[grid["PERMX"] > 0]  # Remove inactive grid cells

    grid["value"] = grid[map_value]
    grid["FLOWI+"] = grid[f"{flow_value}I+"]
    grid["FLOWJ+"] = grid[f"{flow_value}J+"]

    # Webviz map component uses different corner point terminology than libecl
    for (new, old) in [
        ("x0", "x1"),
        ("x1", "x2"),
        ("x2", "x4"),
        ("y0", "y1"),
        ("y1", "y2"),
        ("y2", "y4"),
    ]:
        grid[new] = grid[old]

    return grid


@webvizstore
def get_path(path: Path) -> Path:
    return Path(path)



import base64
import io
import json
import math
import zipfile
from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np
from dash import ALL, MATCH, Input, Output, State, callback, callback_context, no_update
from dash.exceptions import PreventUpdate
from webviz_config import EncodedFile
from webviz_config.utils._dash_component_utils import calculate_slider_step

from webviz_subsurface._providers import (
    EnsembleFaultPolygonsProvider,
    EnsembleSurfaceProvider,
    FaultPolygonsServer,
    ObservedSurfaceAddress,
    QualifiedDiffSurfaceAddress,
    QualifiedSurfaceAddress,
    SimulatedFaultPolygonsAddress,
    SimulatedSurfaceAddress,
    StatisticalSurfaceAddress,
    SurfaceAddress,
    SurfaceArrayServer,
)

from ._layer_model import DeckGLMapLayersModel
from ._tmp_well_pick_provider import WellPickProvider
from ._types import LayerTypes, SurfaceMode
from .layout import (
    DefaultSettings,
    LayoutElements,
    LayoutLabels,
    SideBySideSelectorFlex,
    Tabs,
    update_map_layers,
)


# pylint: disable=too-many-locals,too-many-statements, too-many-arguments
def plugin_callbacks(
    get_uuid: Callable,
    ensemble_surface_providers: Dict[str, EnsembleSurfaceProvider],
    surface_server: SurfaceArrayServer,
    ensemble_fault_polygons_providers: Dict[str, EnsembleFaultPolygonsProvider],
    fault_polygons_server: FaultPolygonsServer,
    map_surface_names_to_fault_polygons: Dict[str, str],
    well_picks_provider: Optional[WellPickProvider],
    fault_polygon_attribute: Optional[str],
    color_tables: List[Dict],
    plugin_data_output: Output,
    plugin_data_requested: Input,
) -> None:
    def selections(tab: str, colorselector: bool = False) -> Dict[str, str]:
        uuid = get_uuid(
            LayoutElements.SELECTIONS
            if not colorselector
            else LayoutElements.COLORSELECTIONS
        )
        return {"view": ALL, "id": uuid, "tab": tab, "selector": ALL}

    def selector_wrapper(tab: str, colorselector: bool = False) -> Dict[str, str]:
        uuid = get_uuid(
            LayoutElements.WRAPPER if not colorselector else LayoutElements.COLORWRAPPER
        )
        return {"id": uuid, "tab": tab, "selector": ALL}

    def links(tab: str, colorselector: bool = False) -> Dict[str, str]:
        uuid = get_uuid(
            LayoutElements.LINK if not colorselector else LayoutElements.COLORLINK
        )
        return {"id": uuid, "tab": tab, "selector": ALL}

    @callback(
        Output(get_uuid(LayoutElements.OPTIONS_DIALOG), "open"),
        Input({"id": get_uuid("Button"), "tab": ALL}, "n_clicks"),
        State(get_uuid(LayoutElements.OPTIONS_DIALOG), "open"),
    )
    def open_close_options_dialog(_n_click: list, is_open: bool) -> bool:
        if any(click is not None for click in _n_click):
            return not is_open
        raise PreventUpdate

    # 2nd callback
    @callback(
        Output({"id": get_uuid(LayoutElements.LINKED_VIEW_DATA), "tab": MATCH}, "data"),
        Output(selector_wrapper(MATCH), "children"),
        Input(selections(MATCH), "value"),
        Input({"id": get_uuid(LayoutElements.VIEWS), "tab": MATCH}, "value"),
        Input({"id": get_uuid(LayoutElements.MULTI), "tab": MATCH}, "value"),
        Input(links(MATCH), "value"),
        Input(get_uuid(LayoutElements.REALIZATIONS_FILTER), "value"),
        Input(get_uuid("tabs"), "value"),
        State(selections(MATCH), "id"),
        State(selector_wrapper(MATCH), "id"),
    )
    def _update_components_and_selected_data(
        mapselector_values: List[Dict[str, Any]],
        number_of_views: int,
        multi_selector: str,
        selectorlinks: List[List[str]],
        filtered_reals: List[int],
        tab_name: str,
        selector_ids: List[dict],
        wrapper_ids: List[dict],
    ) -> Tuple[List[dict], list]:
        """Reads stored raw selections, stores valid selections as a dcc.Store
        and updates visible and valid selections in layout"""

        # Prevent update if the pattern matched components does not match the current tab
        datatab = wrapper_ids[0]["tab"]
        if datatab != tab_name or number_of_views is None:
            raise PreventUpdate

        ctx = callback_context.triggered[0]["prop_id"]

        selector_values = []
        for idx in range(number_of_views):
            view_selections = combine_selector_values_and_name(
                mapselector_values, selector_ids, view=idx
            )
            view_selections["multi"] = multi_selector
            if tab_name == Tabs.STATS:
                view_selections["mode"] = DefaultSettings.VIEW_LAYOUT_STATISTICS_TAB[
                    idx
                ]
            selector_values.append(view_selections)

        linked_selector_names = [l[0] for l in selectorlinks if l]

        component_properties = _update_selector_component_properties_from_provider(
            selector_values=selector_values,
            linked_selectors=linked_selector_names,
            multi=multi_selector,
            multi_in_ctx=get_uuid(LayoutElements.MULTI) in ctx,
            filtered_realizations=filtered_reals,
        )
        # retrive the updated selector values from the component properties
        selector_values = [
            {key: val["value"] for key, val in data.items()}
            for idx, data in enumerate(component_properties)
        ]

        if multi_selector is not None:
            selector_values = update_selections_with_multi(
                selector_values, multi_selector
            )
        selector_values = remove_data_if_not_valid(selector_values)
        if tab_name == Tabs.DIFF and len(selector_values) == 2:
            selector_values = add_diff_surface_to_values(selector_values)

        return (
            selector_values,
            [
                SideBySideSelectorFlex(
                    tab_name,
                    get_uuid,
                    selector=id_val["selector"],
                    view_data=[
                        data[id_val["selector"]] for data in component_properties
                    ],
                    link=id_val["selector"] in linked_selector_names,
                    dropdown=id_val["selector"] in ["ensemble", "mode"],
                )
                for id_val in wrapper_ids
            ],
        )

    # 3rd callback
    @callback(
        Output(
            {"id": get_uuid(LayoutElements.VERIFIED_VIEW_DATA), "tab": MATCH}, "data"
        ),
        Output(selector_wrapper(MATCH, colorselector=True), "children"),
        Input({"id": get_uuid(LayoutElements.LINKED_VIEW_DATA), "tab": MATCH}, "data"),
        Input(selections(MATCH, colorselector=True), "value"),
        Input(
            {"view": ALL, "id": get_uuid(LayoutElements.RANGE_RESET), "tab": MATCH},
            "n_clicks",
        ),
        Input(links(MATCH, colorselector=True), "value"),
        State({"id": get_uuid(LayoutElements.MULTI), "tab": MATCH}, "value"),
        State(selector_wrapper(MATCH, colorselector=True), "id"),
        State(get_uuid(LayoutElements.STORED_COLOR_SETTINGS), "data"),
        State(get_uuid("tabs"), "value"),
        State(selections(MATCH, colorselector=True), "id"),
    )
    def _update_color_components_and_value(
        selector_values: List[dict],
        colorvalues: List[Dict[str, Any]],
        _n_click: int,
        colorlinks: List[List[str]],
        multi: str,
        color_wrapper_ids: List[dict],
        stored_color_settings: Dict,
        tab: str,
        colorval_ids: List[dict],
    ) -> Tuple[List[dict], list]:
        """Adds color settings to validated stored selections, updates color component in layout
        and writes validated selectors with colors to a dcc.Store"""
        ctx = callback_context.triggered[0]["prop_id"]

        if selector_values is None:
            raise PreventUpdate

        reset_color_index = (
            json.loads(ctx.split(".")[0])["view"]
            if get_uuid(LayoutElements.RANGE_RESET) in ctx
            else None
        )
        color_update_index = (
            json.loads(ctx.split(".")[0]).get("view")
            if LayoutElements.COLORSELECTIONS in ctx
            else None
        )

        # if a selector is set as multi in the "Maps per Selector" tab
        # color_range should be linked and the min and max surface ranges
        # from all the views should be used as range
        use_range_from_all = tab == Tabs.SPLIT and multi != "attribute"

        links = [l[0] for l in colorlinks if l]
        links = links if not use_range_from_all else links + ["color_range"]

        # update selector_values with current values from the color components
        for idx, data in enumerate(selector_values):
            data.update(
                combine_selector_values_and_name(colorvalues, colorval_ids, view=idx)
            )
        color_component_properties = _update_color_component_properties(
            values=selector_values,
            links=links if not use_range_from_all else links + ["color_range"],
            stored_color_settings=stored_color_settings,
            reset_color_index=reset_color_index,
            color_update_index=color_update_index,
        )

        if use_range_from_all:
            ranges = [data["surface_range"] for data in selector_values]
            min_max_for_all = [min(r[0] for r in ranges), max(r[1] for r in ranges)]
            for data in color_component_properties:
                data["color_range"]["range"] = min_max_for_all
                if reset_color_index is not None:
                    data["color_range"]["value"] = min_max_for_all

        for idx, data in enumerate(color_component_properties):
            for key, val in data.items():
                selector_values[idx][key] = val["value"]

        return (
            selector_values,
            [
                SideBySideSelectorFlex(
                    tab,
                    get_uuid,
                    selector=id_val["selector"],
                    view_data=[
                        data[id_val["selector"]] for data in color_component_properties
                    ],
                    link=id_val["selector"] in links,
                    dropdown=id_val["selector"] in ["colormap"],
                )
                for id_val in color_wrapper_ids
            ],
        )

    # 4th callback
    @callback(
        Output(get_uuid(LayoutElements.STORED_COLOR_SETTINGS), "data"),
        Input({"id": get_uuid(LayoutElements.VERIFIED_VIEW_DATA), "tab": ALL}, "data"),
        State(get_uuid("tabs"), "value"),
        State(get_uuid(LayoutElements.STORED_COLOR_SETTINGS), "data"),
        State({"id": get_uuid(LayoutElements.VERIFIED_VIEW_DATA), "tab": ALL}, "id"),
    )
    def _update_color_store(
        selector_values: List[List[dict]],
        tab: str,
        stored_color_settings: Dict[str, dict],
        data_id: List[dict],
    ) -> Dict[str, dict]:
        """Update the color store with chosen color range and colormap for surfaces"""
        if selector_values is None:
            raise PreventUpdate
        index = [x["tab"] for x in data_id].index(tab)

        stored_color_settings = (
            stored_color_settings if stored_color_settings is not None else {}
        )
        for data in selector_values[index]:
            surfaceid = (
                get_surface_id_for_diff_surf(selector_values[index])
                if data.get("surf_type") == "diff"
                else get_surface_id_from_data(data)
            )
            stored_color_settings[surfaceid] = {
                "colormap": data["colormap"],
                "color_range": data["color_range"],
            }

        return stored_color_settings

    # 5th callback
    @callback(
        Output({"id": get_uuid(LayoutElements.DECKGLMAP), "tab": MATCH}, "layers"),
        # Output({"id": get_uuid(LayoutElements.DECKGLMAP), "tab": MATCH}, "bounds"),
        Output({"id": get_uuid(LayoutElements.DECKGLMAP), "tab": MATCH}, "views"),
        Input(
            {"id": get_uuid(LayoutElements.VERIFIED_VIEW_DATA), "tab": MATCH}, "data"
        ),
        Input(get_uuid(LayoutElements.WELLS), "value"),
        Input(get_uuid(LayoutElements.VIEW_COLUMNS), "value"),
        Input(get_uuid(LayoutElements.OPTIONS), "value"),
        State(get_uuid("tabs"), "value"),
        State({"id": get_uuid(LayoutElements.MULTI), "tab": MATCH}, "value"),
    )
    def _update_map(
        surface_elements: List[dict],
        selected_wells: List[str],
        view_columns: Optional[int],
        options: List[str],
        tab_name: str,
        multi: str,
    ) -> tuple:
        """Updates the map component with the stored, validated selections"""

        # Prevent update if the pattern matched components does not match the current tab
        state_list = callback_context.states_list
        if state_list[1]["id"]["tab"] != tab_name or surface_elements is None:
            raise PreventUpdate

        if tab_name == Tabs.DIFF:
            view_columns = 3 if view_columns is None else view_columns

        layers = update_map_layers(
            views=len(surface_elements),
            include_well_layer=well_picks_provider is not None,
            visible_well_layer=LayoutLabels.SHOW_WELLS in options,
            visible_fault_polygons_layer=LayoutLabels.SHOW_FAULTPOLYGONS in options,
        )
        layer_model = DeckGLMapLayersModel(layers)

        for idx, data in enumerate(surface_elements):
            diff_surf = data.get("surf_type") == "diff"
            surf_meta, mesh_url = (
                get_surface_metadata_and_image(data)
                if not diff_surf
                else get_surface_metadata_and_image_for_diff_surface(surface_elements)
            )
            if (
                data["color_range"][0] != surf_meta.val_min
                or data["color_range"][1] != surf_meta.val_max
            ):
                color_range = data["color_range"]
            else:
                color_range = None
            layer_data = {
                "meshUrl": mesh_url,
                "frame": {
                    "origin": [surf_meta.x_ori, surf_meta.y_ori],
                    "count": [surf_meta.x_count, surf_meta.y_count],
                    "increment": [surf_meta.x_inc, surf_meta.y_inc],
                    "rotDeg": surf_meta.rot_deg,
                },
                "colorMapName": data["colormap"],
                "colorMapRange": color_range,
            }
            layer_idx = None
            for layer in layers:
                if layer["id"] == f"{LayoutElements.MAP3D_LAYER}-{idx}":
                    layer_idx = layers.index(layer)
                    break
            if layer_idx is not None:
                layers[layer_idx].update(layer_data)
            else:
                layer_data["id"] = f"{LayoutElements.MAP3D_LAYER}-{idx}"
                layer_data["@@type"] = LayerTypes.MAP3D
                layer_data["material"] = False
                layers.insert(0, layer_data)

            if (
                LayoutLabels.SHOW_FAULTPOLYGONS in options
                and fault_polygon_attribute is not None
            ):
                # if diff surface use polygons from first view
                data_for_faultpolygons = data if not diff_surf else surface_elements[0]
                fault_polygons_provider = ensemble_fault_polygons_providers[
                    data_for_faultpolygons["ensemble"][0]
                ]
                horizon_name = data_for_faultpolygons["name"][0]
                fault_polygons_address = SimulatedFaultPolygonsAddress(
                    attribute=fault_polygon_attribute,
                    name=map_surface_names_to_fault_polygons.get(
                        horizon_name, horizon_name
                    ),
                    realization=int(data_for_faultpolygons["realizations"][0]),
                )
                layer_model.update_layer_by_id(
                    layer_id=f"{LayoutElements.FAULTPOLYGONS_LAYER}-{idx}",
                    layer_data={
                        "data": fault_polygons_server.encode_partial_url(
                            provider_id=fault_polygons_provider.provider_id(),
                            fault_polygons_address=fault_polygons_address,
                        ),
                    },
                )
            if LayoutLabels.SHOW_WELLS in options and well_picks_provider is not None:
                horizon_name = (
                    data["name"][0] if not diff_surf else surface_elements[0]["name"][0]
                )
                layer_model.update_layer_by_id(
                    layer_id=f"{LayoutElements.WELLS_LAYER}-{idx}",
                    layer_data={
                        "data": well_picks_provider.get_geojson(
                            selected_wells, horizon_name
                        )
                    },
                )
        return (
            layer_model.layers,
            # viewport_bounds if surface_elements else no_update,
            {
                "layout": view_layout(len(surface_elements), view_columns),
                "showLabel": True,
                "viewports": [
                    {
                        "id": f"{view}_view",
                        "show3D": False,
                        "isSync": True,
                        "layerIds": [
                            f"{LayoutElements.MAP3D_LAYER}-{view}",
                            f"{LayoutElements.FAULTPOLYGONS_LAYER}-{view}",
                            f"{LayoutElements.WELLS_LAYER}-{view}",
                        ],
                        "name": make_viewport_label(
                            surface_elements[view], tab_name, multi
                        ),
                    }
                    for view in range(len(surface_elements))
                ],
            },
        )

    @callback(
        plugin_data_output,
        plugin_data_requested,
        State(
            {"id": get_uuid(LayoutElements.VERIFIED_VIEW_DATA), "tab": ALL},
            "data",
        ),
        State(get_uuid("tabs"), "value"),
        State({"id": get_uuid(LayoutElements.VERIFIED_VIEW_DATA), "tab": ALL}, "id"),
        prevent_initial_call=True,
    )
    def _user_download_data(
        data_requested: bool,
        surface_elements_in_tabs: List[List[Dict]],
        tab_name: str,
        tab_ids: List[dict],
    ) -> Optional[EncodedFile]:
        """Callback for downloading surfaces from the plugin"""
        if not surface_elements_in_tabs or not data_requested:
            return no_update
        index = [x["tab"] for x in tab_ids].index(tab_name)
        surface_elements_in_tab = surface_elements_in_tabs[index]
        surface_bytes = {}
        reals_in_statistics = []
        for s_elem in surface_elements_in_tab:
            if s_elem.get("surf_type") == "diff":
                continue
            surface_address = get_surface_address_from_data(s_elem)
            ensemble = s_elem["ensemble"][0]
            provider = ensemble_surface_providers[ensemble]
            surface = provider.get_surface(surface_address)
            if surface is None:
                continue
            date_str = s_elem["date"][0] if s_elem["date"] else ""
            realizations = s_elem["realizations"]
            mode_str = (
                s_elem["mode"]
                if SurfaceMode(s_elem["mode"]) != SurfaceMode.REALIZATION
                else realizations[0]
            )
            surface_name = (
                f"{s_elem['attribute'][0]}--{s_elem['name'][0]}"
                f"--{date_str}--{mode_str}--{ensemble}.gri"
            )
            if SurfaceMode(s_elem["mode"]) not in [
                SurfaceMode.OBSERVED,
                SurfaceMode.REALIZATION,
            ]:
                reals_in_statistics.append(
                    f"{surface_name},"
                    f"INCLUDED REALIZATIONS: {realizations}, EXCLUDED REALIZATIONS: "
                    f"{[real for real in provider.realizations() if real not in realizations]}"
                )
            byte_stream = io.BytesIO()
            surface.to_file(byte_stream)
            surface_bytes[surface_name] = byte_stream

        with io.BytesIO() as bytes_io:
            with zipfile.ZipFile(bytes_io, "w") as zipped_data:
                for surface_name, surface_byte in surface_bytes.items():
                    zipped_data.writestr(surface_name, surface_byte.getvalue())
                zipped_data.writestr(
                    "statistics.txt",
                    "\n".join(str(real) for real in reals_in_statistics),
                )
            return {
                "filename": "surfaces.zip",
                "content": base64.b64encode(bytes_io.getvalue()).decode("ascii"),
                "mime_type": "application/zip",
            }

    def make_viewport_label(data: dict, tab: str, multi: Optional[str]) -> str:
        """Return text-label for each viewport based on which tab is selected"""
        # For the difference view
        if tab == Tabs.DIFF and data.get("surf_type") == "diff":
            return "Difference Map (View1 - View2)"

        # For the statistics view 'mode' is used as label
        if tab == Tabs.STATS:
            if data["mode"] == SurfaceMode.REALIZATION:
                return f"REAL {data['realizations'][0]}"
            return data["mode"]

        # For the "map per selector" the chosen multi selector is used as label
        if tab == Tabs.SPLIT:
            if multi == "realizations":
                return f"REAL {data['realizations'][0]}"
            return data[multi][0] if multi == "mode" else data[multi]

        return general_label(data)

    def general_label(data: dict) -> str:
        """Create a general map label with all available information"""
        surfaceid = [data["ensemble"][0], data["attribute"][0], data["name"][0]]
        if data["date"]:
            surfaceid.append(data["date"][0])
        if data["mode"] != SurfaceMode.REALIZATION:
            surfaceid.append(data["mode"])
        else:
            surfaceid.append(f"REAL {data['realizations'][0]}")
        return " ".join(surfaceid)

    # pylint: disable=too-many-branches
    def _update_selector_component_properties_from_provider(
        selector_values: List[dict],
        linked_selectors: List[str],
        multi: str,
        multi_in_ctx: bool,
        filtered_realizations: List[int],
    ) -> List[Dict[str, dict]]:
        """Return updated options and values for the different selector components using
        the provider. If current selected value for a selector is in the options it will
        be used as the new value"""
        view_data = []
        for idx, data in enumerate(selector_values):

            if not ("ensemble" in linked_selectors and idx > 0):
                ensembles = list(ensemble_surface_providers.keys())
                ensemble = data.get("ensemble", [])
                ensemble = [ensemble] if isinstance(ensemble, str) else ensemble
                if not ensemble or multi_in_ctx:
                    ensemble = ensembles if multi == "ensemble" else ensembles[:1]

            if not ("attribute" in linked_selectors and idx > 0):
                attributes = []
                for ens in ensemble:
                    provider = ensemble_surface_providers[ens]
                    attributes.extend(
                        [x for x in provider.attributes() if x not in attributes]
                    )
                    # only show attributes with date when multi is set to date
                    if multi == "date":
                        attributes = [
                            x for x in attributes if attribute_has_date(x, provider)
                        ]

                attribute = [x for x in data.get("attribute", []) if x in attributes]
                if not attribute or multi_in_ctx:
                    attribute = attributes if multi == "attribute" else attributes[:1]

            if not ("name" in linked_selectors and idx > 0):
                names = []
                for ens in ensemble:
                    provider = ensemble_surface_providers[ens]
                    for attr in attribute:
                        attr_names = provider.surface_names_for_attribute(attr)
                        names.extend([x for x in attr_names if x not in names])

                name = [x for x in data.get("name", []) if x in names]
                if not name or multi_in_ctx:
                    name = names if multi == "name" else names[:1]

            if not ("date" in linked_selectors and idx > 0):
                dates = []
                for ens in ensemble:
                    provider = ensemble_surface_providers[ens]
                    for attr in attribute:
                        attr_dates = provider.surface_dates_for_attribute(attr)

                        if attr_dates is not None:
                            dates.extend([x for x in attr_dates if x not in dates])

                interval_dates = [x for x in dates if "_" in x]
                dates = [x for x in dates if x not in interval_dates] + interval_dates

                date = [x for x in data.get("date", []) if x in dates]
                if not date or multi_in_ctx:
                    date = dates if multi == "date" else dates[:1]

            if not ("mode" in linked_selectors and idx > 0):
                modes = list(SurfaceMode)
                mode = data.get("mode", SurfaceMode.REALIZATION)

            if not ("realizations" in linked_selectors and idx > 0):
                if mode == SurfaceMode.REALIZATION:
                    real = data.get("realizations", [])
                    if not real or real[0] not in filtered_realizations:
                        real = filtered_realizations[:1]
                else:
                    real = filtered_realizations

            view_data.append(
                {
                    "ensemble": {
                        "value": ensemble,
                        "options": ensembles,
                        "multi": multi == "ensemble",
                    },
                    "attribute": {
                        "value": attribute,
                        "options": attributes,
                        "multi": multi == "attribute",
                    },
                    "name": {"value": name, "options": names, "multi": multi == "name"},
                    "date": {"value": date, "options": dates, "multi": multi == "date"},
                    "mode": {"value": mode, "options": modes},
                    "realizations": {
                        "value": real,
                        "disabled": mode != SurfaceMode.REALIZATION,
                        "options": filtered_realizations,
                        "multi": mode != SurfaceMode.REALIZATION
                        or multi == "realizations",
                    },
                }
            )

        return view_data

    def _update_color_component_properties(
        values: List[dict],
        links: List[str],
        stored_color_settings: dict,
        reset_color_index: Optional[int],
        color_update_index: Optional[int],
    ) -> List[dict]:
        """Return updated options and values for the different color selector components.
        If previous color settings are found it will be used, or set value to min and max
        surface range unless the user has updated the component through interaction."""
        stored_color_settings = (
            stored_color_settings if stored_color_settings is not None else {}
        )
        colormaps = [color_table["name"] for color_table in color_tables]

        surfids: List[str] = []
        color_data: List[dict] = []
        for idx, data in enumerate(values):
            surfaceid = (
                get_surface_id_for_diff_surf(values)
                if data.get("surf_type") == "diff"
                else get_surface_id_from_data(data)
            )
            # if surfaceid exist in another view use the color settings
            # from that view and disable the color component
            if surfaceid in surfids:
                index_of_first = surfids.index(surfaceid)
                surfids.append(surfaceid)
                view_data = deepcopy(color_data[index_of_first])
                view_data["colormap"].update(disabled=True)
                view_data["color_range"].update(disabled=True)
                color_data.append(view_data)
                continue

            surfids.append(surfaceid)

            use_stored_color = (
                surfaceid in stored_color_settings and not color_update_index == idx
            )
            if not ("colormap" in links and idx > 0):
                colormap = (
                    stored_color_settings[surfaceid]["colormap"]
                    if use_stored_color
                    else data.get("colormap", colormaps[0])
                )

            if not ("color_range" in links and idx > 0):
                value_range = data["surface_range"]
                if data.get("colormap_keep_range", False):
                    color_range = data["color_range"]
                elif reset_color_index == idx or surfaceid not in stored_color_settings:
                    color_range = value_range
                else:
                    color_range = (
                        stored_color_settings[surfaceid]["color_range"]
                        if use_stored_color
                        else data.get("color_range", value_range)
                    )

            color_data.append(
                {
                    "colormap": {"value": colormap, "options": colormaps},
                    "color_range": {
                        "value": color_range,
                        "step": calculate_slider_step(
                            min_value=value_range[0],
                            max_value=value_range[1],
                            steps=100,
                        )
                        if value_range[0] != value_range[1]
                        else 0,
                        "range": value_range,
                    },
                }
            )

        return color_data

    def get_surface_address_from_data(
        data: dict,
    ) -> Union[
        SimulatedSurfaceAddress, ObservedSurfaceAddress, StatisticalSurfaceAddress
    ]:
        """Return the SurfaceAddress based on view selection"""
        has_date = (
            ensemble_surface_providers[data["ensemble"][0]].surface_dates_for_attribute(
                data["attribute"][0]
            )
            is not None
        )

        if data["mode"] == SurfaceMode.REALIZATION:
            return SimulatedSurfaceAddress(
                attribute=data["attribute"][0],
                name=data["name"][0],
                datestr=data["date"][0] if has_date else None,
                realization=int(data["realizations"][0]),
            )
        if data["mode"] == SurfaceMode.OBSERVED:
            return ObservedSurfaceAddress(
                attribute=data["attribute"][0],
                name=data["name"][0],
                datestr=data["date"][0] if has_date else None,
            )
        return StatisticalSurfaceAddress(
            attribute=data["attribute"][0],
            name=data["name"][0],
            datestr=data["date"][0] if has_date else None,
            realizations=[int(real) for real in data["realizations"]],
            statistic=data["mode"],
        )

    def publish_and_get_surface_metadata(
        surface_provider: EnsembleSurfaceProvider, surface_address: SurfaceAddress
    ) -> Tuple:
        provider_id: str = surface_provider.provider_id()
        qualified_address = QualifiedSurfaceAddress(provider_id, surface_address)
        surf_meta = surface_server.get_surface_metadata(qualified_address)
        if not surf_meta:
            # This means we need to compute the surface
            surface = surface_provider.get_surface(address=surface_address)
            if not surface:
                raise ValueError(
                    f"Could not get surface for address: {surface_address}"
                )
            surface_server.publish_surface(qualified_address, surface)
            surf_meta = surface_server.get_surface_metadata(qualified_address)
        return surf_meta, surface_server.encode_partial_url(qualified_address)

    def publish_and_get_diff_surface_metadata(
        surface_provider: EnsembleSurfaceProvider,
        surface_address: SurfaceAddress,
        sub_surface_provider: EnsembleSurfaceProvider,
        sub_surface_address: SurfaceAddress,
    ) -> Tuple:
        provider_id: str = surface_provider.provider_id()
        subprovider_id = sub_surface_provider.provider_id()
        qualified_address: Union[QualifiedSurfaceAddress, QualifiedDiffSurfaceAddress]
        qualified_address = QualifiedDiffSurfaceAddress(
            provider_id, surface_address, subprovider_id, sub_surface_address
        )

        surf_meta = surface_server.get_surface_metadata(qualified_address)
        if not surf_meta:
            surface_a = surface_provider.get_surface(address=surface_address)
            surface_b = sub_surface_provider.get_surface(address=sub_surface_address)
            if surface_a is not None and surface_b is not None:
                surface = surface_a - surface_b
            surface_server.publish_surface(qualified_address, surface)
            surf_meta = surface_server.get_surface_metadata(qualified_address)
        return surf_meta, surface_server.encode_partial_url(qualified_address)

    def get_surface_id_from_data(data: dict) -> str:
        """Retrieve surfaceid used for the colorstore"""
        surfaceid = data["attribute"][0] + data["name"][0]
        if data["date"]:
            surfaceid += data["date"][0]
        if data["mode"] == SurfaceMode.STDDEV:
            surfaceid += data["mode"]
        return surfaceid

    def get_surface_id_for_diff_surf(values: List[dict]) -> str:
        """Retrieve surfaceid used for the colorstore, for the diff surface
        this needs to be a combination of the two surfaces subtracted"""
        return get_surface_id_from_data(values[0]) + get_surface_id_from_data(values[1])

    def update_selections_with_multi(values: List[dict], multi: str) -> List[dict]:
        """If a selector has been set as multi, the values selected in that component needs
        to be divided between the views so that there is only one unique value in each"""
        multi_values = values[0][multi]
        new_values = []
        for val in multi_values:
            updated_values = deepcopy(values[0])
            updated_values[multi] = [val]
            new_values.append(updated_values)
        return new_values

    def attribute_has_date(attribute: str, provider: EnsembleSurfaceProvider) -> bool:
        """Check if an attribute has any dates"""
        return provider.surface_dates_for_attribute(attribute) is not None

    def remove_data_if_not_valid(values: List[dict]) -> List[dict]:
        """Checks if surfaces can be provided from the selections.
        Any invalid selections are removed."""
        updated_values = []
        for data in values:
            surface_address = get_surface_address_from_data(data)
            try:
                provider = ensemble_surface_providers[data["ensemble"][0]]
                surf_meta, _ = publish_and_get_surface_metadata(
                    surface_address=surface_address,
                    surface_provider=provider,
                )
            except ValueError:
                continue
            if not isinstance(
                surf_meta.val_min, np.ma.core.MaskedConstant
            ) and not isinstance(surf_meta.val_max, np.ma.core.MaskedConstant):
                data["surface_range"] = [surf_meta.val_min, surf_meta.val_max]
                updated_values.append(data)

        return updated_values

    def get_surface_metadata_and_image(data: dict) -> Tuple:
        surface_address = get_surface_address_from_data(data)
        provider = ensemble_surface_providers[data["ensemble"][0]]
        return publish_and_get_surface_metadata(
            surface_address=surface_address, surface_provider=provider
        )

    def get_surface_metadata_and_image_for_diff_surface(
        selector_values: List[dict],
    ) -> Tuple:
        surface_address = get_surface_address_from_data(selector_values[0])
        sub_surface_address = get_surface_address_from_data(selector_values[1])
        provider = ensemble_surface_providers[selector_values[0]["ensemble"][0]]
        sub_provider = ensemble_surface_providers[selector_values[1]["ensemble"][0]]
        return publish_and_get_diff_surface_metadata(
            surface_address=surface_address,
            surface_provider=provider,
            sub_surface_address=sub_surface_address,
            sub_surface_provider=sub_provider,
        )

    def add_diff_surface_to_values(selector_values: List[dict]) -> List[dict]:
        surf_meta, _ = get_surface_metadata_and_image_for_diff_surface(selector_values)
        selector_values.append(
            {
                "surface_range": [surf_meta.val_min, surf_meta.val_max],
                "surf_type": "diff",
            }
        )
        return selector_values

    def combine_selector_values_and_name(
        values: list, id_list: list, view: int
    ) -> dict:
        """Combine selector values with selector name for given view"""
        return {
            id_values["selector"]: val
            for val, id_values in zip(values, id_list)
            if id_values["view"] == view
        }


def view_layout(views: int, columns: Optional[int] = None) -> List[int]:
    """Function to set number of rows and columns for the map, if number
    of columns is not specified, a square matrix layout is used"""
    columns = (
        columns
        if columns is not None
        else min([x for x in range(1, 20, 1) if (x * x) >= views])
    )
    rows = math.ceil(views / columns)
    return [rows, columns]



import copy
from typing import Any, Dict, List

from webviz_subsurface.plugins._map_viewer_fmu.color_tables import default_color_tables

ColorTables = List[Dict[str, Any]]


def co2leakage_color_tables() -> ColorTables:
    tables = default_color_tables + uniform_color_tables()
    rev_tables = _reversed_color_tables(tables)
    return [t for tup in zip(tables, rev_tables) for t in tup]


def uniform_color_tables() -> ColorTables:
    # Source: https://waldyrious.net/viridis-palette-generator/ + matplotlib._cm_listed
    return [
        {
            "name": "Viridis",
            "discrete": False,
            "colors": [
                [0.0, 253, 231, 37],
                [0.25, 94, 201, 98],
                [0.50, 33, 145, 140],
                [0.75, 59, 82, 139],
                [1.0, 68, 1, 84],
            ],
        },
        {
            "name": "Inferno",
            "discrete": False,
            "colors": [
                [0.0, 252, 255, 164],
                [0.25, 249, 142, 9],
                [0.5, 188, 55, 84],
                [0.75, 87, 16, 110],
                [1.0, 0, 0, 4],
            ],
        },
        {
            "name": "Magma",
            "discrete": False,
            "colors": [
                [0.0, 252, 253, 191],
                [0.25, 252, 137, 97],
                [0.5, 183, 55, 121],
                [0.75, 81, 18, 124],
                [1.0, 0, 0, 4],
            ],
        },
        {
            "name": "Plasma",
            "discrete": False,
            "colors": [
                [0.0, 240, 249, 33],
                [0.25, 248, 149, 64],
                [0.5, 204, 71, 120],
                [0.75, 126, 3, 168],
                [1.0, 13, 8, 135],
            ],
        },
        {
            "name": "Cividis",
            "discrete": False,
            "colors": [
                [0.0, 0, 32, 77],
                [0.25, 64, 77, 107],
                [0.5, 124, 123, 120],
                [0.75, 188, 175, 111],
                [1.0, 255, 234, 70],
            ],
        },
    ]


def _reversed_color_tables(tables: ColorTables) -> ColorTables:
    rev = []
    for _t in tables:
        _new_t = copy.deepcopy(_t)
        _new_t["name"] = _t["name"] + " (Reverse)"
        _new_t["colors"] = [
            [c0[0]] + c1[1:] for c0, c1 in zip(_t["colors"], _t["colors"][::-1])
        ]
        rev.append(_new_t)
    return rev



from pathlib import Path
from typing import List

import pandas as pd


class FilterTable:
    """Uses paths to csv files to render dataframes
    and to update them.
    * `target_points`: File path to targetpoints.csv
    * `well_points`: File path to wellpoints.csv
    """

    def __init__(
        self,
        target_points: Path = None,
        well_points: Path = None,
    ):
        self.target_points = target_points
        self.well_points = well_points

    def get_targetpoints_df(self) -> pd.DataFrame:
        df = pd.read_csv(self.target_points)
        return df.round(2)

    def get_wellpoints_df(self) -> pd.DataFrame:
        df = pd.read_csv(self.well_points)
        return df.round(2)

    def update_wellpoints_df(self, column_list: List[str]) -> pd.DataFrame:
        df = pd.read_csv(self.well_points)
        sorted_list = []
        for colm in df.keys().values:
            for col in column_list:
                if colm == col:
                    sorted_list.append(col)
        return df[sorted_list].round(2)



from webviz_subsurface._utils.formatting import printable_int_list


def test_printable_int_list() -> None:
    assert printable_int_list([1, 2, 5, 6, 9, 8, 19]) == "1-2, 5-6, 8-9, 19"
    assert printable_int_list([5, 3, 4, 0]) == "0, 3-5"
    assert printable_int_list([0, 1, 2, 3, 6]) == "0-3, 6"
    assert printable_int_list([]) == "None"
    assert printable_int_list(None) == "None"
    assert printable_int_list([5]) == "5"



from pathlib import Path
from typing import Dict, List

import pandas as pd
from webviz_config import WebvizPluginABC, WebvizSettings

from webviz_subsurface._providers import EnsembleSummaryProvider
from webviz_subsurface._utils.ensemble_summary_provider_set_factory import (
    create_lazy_ensemble_summary_provider_set_from_paths,
)

from ._plugin_ids import PluginIds
from .shared_settings import Filter, ViewSettings
from .view_elements import Graph
from .views import BarView, FanView, LineView


class BhpQc(WebvizPluginABC):
    """QC simulated bottom hole pressures (BHP) from reservoir simulations.

    Can be used to check if your simulated BHPs are in a realistic range.
    E.g. check if your simulated bottom hole pressures are very low in producers,
    or very high injectors.
    ---

    * **`ensembles`:** Which ensembles in `shared_settings` to visualize.
    * **`rel_file_pattern`:** path to `.arrow` files with summary data.
    ---
    Data is read directly from the arrow files with the raw frequency (not resampled).
    Resampling and csvs are not supported to avoid potential of interpolation, which
    might cover extreme BHP values.

    """

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        ensembles: list,
        rel_file_pattern: str = "share/results/unsmry/*.arrow",
        wells: List[str] = None,
    ):
        super().__init__()

        if ensembles is None:
            raise ValueError("Enembles needs to be provided")

        self.ens_paths: Dict[str, Path] = {
            ensemble_name: webviz_settings.shared_settings["scratch_ensembles"][
                ensemble_name
            ]
            for ensemble_name in ensembles
        }

        self._input_provider_set = create_lazy_ensemble_summary_provider_set_from_paths(
            self.ens_paths,
            rel_file_pattern,
        )

        dfs = []
        column_keys = {}
        for ens_name in ensembles:
            ens_provider = self._input_provider_set.provider(ens_name)
            column_keys[ens_name] = _get_wbhp_vectors(ens_provider, wells)
            df = ens_provider.get_vectors_df(column_keys[ens_name], None)
            df["ENSEMBLE"] = ens_name
            dfs.append(df.loc[:, (df != 0).any(axis=0)])  # remove zero-columns

        self.smry = pd.concat(dfs)
        self.theme = webviz_settings.theme

        self.add_store(
            PluginIds.Stores.SELECTED_ENSEMBLE, WebvizPluginABC.StorageType.SESSION
        )
        self.add_store(
            PluginIds.Stores.SELECTED_SORT_BY, WebvizPluginABC.StorageType.SESSION
        )
        self.add_store(
            PluginIds.Stores.SELECTED_ASCENDING_DESCENDING,
            WebvizPluginABC.StorageType.SESSION,
        )
        self.add_store(
            PluginIds.Stores.SELECTED_MAX_NUMBER_OF_WELLS,
            WebvizPluginABC.StorageType.SESSION,
        )
        self.add_store(
            PluginIds.Stores.SELECTED_WELLS, WebvizPluginABC.StorageType.SESSION
        )
        self.add_store(
            PluginIds.Stores.SELECTED_STATISTICS, WebvizPluginABC.StorageType.SESSION
        )

        self.add_view(
            LineView(self.smry, webviz_settings),
            PluginIds.BhpId.LINE_CHART,
        )
        self.add_view(
            FanView(self.smry, webviz_settings),
            PluginIds.BhpId.FAN_CHART,
        )
        self.add_view(
            BarView(self.smry, webviz_settings),
            PluginIds.BhpId.BAR_CHART,
        )

        self.add_shared_settings_group(
            Filter(self.smry), PluginIds.SharedSettings.FILTER
        )
        self.add_shared_settings_group(
            ViewSettings(),
            PluginIds.SharedSettings.VIEW_SETTINGS,
            [
                self.view(PluginIds.BhpId.BAR_CHART).get_unique_id().to_string(),
                self.view(PluginIds.BhpId.LINE_CHART).get_unique_id().to_string(),
            ],
        )

        # set start-up view (only affects first time use)
        self.set_active_view_id(PluginIds.BhpId.LINE_CHART)

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.view(PluginIds.BhpId.FAN_CHART)
                .view_element(FanView.Ids.FAN_CHART)
                .component_unique_id(Graph.Ids.GRAPH),
                "content": (
                    "Dashboard for BHP QC: "
                    "Check that simulated bottom hole pressures are realistic."
                    " Can be viewed in a Fan chart,"
                ),
            },
            {
                "id": self.view(PluginIds.BhpId.LINE_CHART)
                .view_element(LineView.Ids.LINE_CHART)
                .component_unique_id(Graph.Ids.GRAPH),
                "content": ("Line chart"),
            },
            {
                "id": self.view(PluginIds.BhpId.BAR_CHART)
                .view_element(BarView.Ids.BAR_CHART)
                .component_unique_id(Graph.Ids.GRAPH),
                "content": ("and Bar chart."),
            },
            {
                "id": self.shared_settings_group(
                    PluginIds.SharedSettings.FILTER
                ).component_unique_id(Filter.Ids.ENSEMBLE),
                "content": "Select ensemble to QC.",
            },
            {
                "id": self.shared_settings_group(
                    PluginIds.SharedSettings.FILTER
                ).component_unique_id(Filter.Ids.SORT_BY),
                "content": "Sort wells left to right according to this value.",
            },
            {
                "id": self.shared_settings_group(
                    PluginIds.SharedSettings.FILTER
                ).component_unique_id(Filter.Ids.MAX_NUMBER_OF_WELLS_SLIDER),
                "content": (
                    "Show max selected number of top ranked wells after sorting and filtering."
                ),
            },
            {
                "id": self.shared_settings_group(
                    PluginIds.SharedSettings.FILTER
                ).component_unique_id(Filter.Ids.WELLS),
                "content": "Filter wells.",
            },
        ]


# ---------------------------
def _get_wbhp_vectors(
    ens_provider: EnsembleSummaryProvider,
    wells: List[str] = None,
) -> list:
    """Return list of WBHP vectors. If wells arg is None, return for all wells."""

    if wells is not None:
        return [f"WBHP:{well}" for well in wells]

    wbhp_vectors = [
        vector for vector in ens_provider.vector_names() if vector.startswith("WBHP:")
    ]
    if not wbhp_vectors:
        raise RuntimeError("No WBHP vectors found.")

    return wbhp_vectors



from typing import Any, Dict, List, Optional, Tuple

import webviz_core_components as wcc
from dash import (
    ALL,
    MATCH,
    Input,
    Output,
    callback,
    callback_context,
    dcc,
    html,
    no_update,
)
from dash.development.base_component import Component
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from webviz_subsurface._providers.ensemble_grid_provider import (
    CellFilter,
    EnsembleGridProvider,
)
from webviz_subsurface.plugins._grid_viewer_fmu._layout_elements import ElementIds
from webviz_subsurface.plugins._grid_viewer_fmu._types import GRIDDIRECTION


def list_to_options(values: List) -> List:
    return [{"value": val, "label": val} for val in values]


class GridFilter(SettingsGroupABC):
    def __init__(
        self, grid_provider: EnsembleGridProvider, initial_grid_filter: Dict[str, int]
    ) -> None:
        super().__init__("Grid IJK Filter")
        self.grid_provider = grid_provider
        initial_grid = grid_provider.get_3dgrid(grid_provider.realizations()[0])
        self.initial_grid_filter = initial_grid_filter
        self.grid_dimensions = (
            CellFilter(
                i_min=1,
                j_min=1,
                k_min=1,
                i_max=initial_grid.dimensions[0],
                j_max=initial_grid.dimensions[1],
                k_max=initial_grid.dimensions[2],
            )
            if initial_grid
            else CellFilter(i_min=1, j_min=1, k_min=1, i_max=1, j_max=1, k_max=1)
        )

    def layout(self) -> List[Component]:

        return [
            html.Div(
                style={"backgroundColor": "white"},
                children=[
                    crop_widget(
                        widget_id=self.get_unique_id().to_string(),
                        min_val=self.grid_dimensions.i_min,
                        selected_min_val=self.initial_grid_filter.get("i_start", None),
                        selected_max_val=self.initial_grid_filter.get("i_width", None),
                        max_val=self.grid_dimensions.i_max,
                        direction=GRIDDIRECTION.I,
                    ),
                    crop_widget(
                        widget_id=self.get_unique_id().to_string(),
                        min_val=self.grid_dimensions.j_min,
                        selected_min_val=self.initial_grid_filter.get("j_start", None),
                        selected_max_val=self.initial_grid_filter.get("j_width", None),
                        max_val=self.grid_dimensions.j_max,
                        direction=GRIDDIRECTION.J,
                    ),
                    crop_widget(
                        widget_id=self.get_unique_id().to_string(),
                        min_val=self.grid_dimensions.k_min,
                        selected_min_val=self.initial_grid_filter.get("k_start", None),
                        selected_max_val=self.initial_grid_filter.get("k_width", None),
                        max_val=self.grid_dimensions.k_max,
                        direction=GRIDDIRECTION.K,
                    ),
                ],
            ),
        ]

    def set_callbacks(self) -> None:
        @callback(
            Output(
                {
                    "id": self.get_unique_id().to_string(),
                    "direction": MATCH,
                    "component": "input",
                    "component2": MATCH,
                },
                "value",
            ),
            Output(
                {
                    "id": self.get_unique_id().to_string(),
                    "direction": MATCH,
                    "component": "slider",
                    "component2": MATCH,
                },
                "value",
            ),
            Input(
                {
                    "id": self.get_unique_id().to_string(),
                    "direction": MATCH,
                    "component": "input",
                    "component2": MATCH,
                },
                "value",
            ),
            Input(
                {
                    "id": self.get_unique_id().to_string(),
                    "direction": MATCH,
                    "component": "slider",
                    "component2": MATCH,
                },
                "value",
            ),
        )
        def _synchronize_crop_slider_and_input(
            input_val: int, slider_val: int
        ) -> Tuple[Any, Any]:
            trigger_id = callback_context.triggered[0]["prop_id"].split(".")[0]
            if "slider" in trigger_id:
                return slider_val, no_update
            return no_update, input_val

        @callback(
            Output(self.get_store_unique_id(ElementIds.IJK_CROP_STORE), "data"),
            Input(
                {
                    "id": self.get_unique_id().to_string(),
                    "direction": ALL,
                    "component": "input",
                    "component2": "start",
                },
                "value",
            ),
            Input(
                {
                    "id": self.get_unique_id().to_string(),
                    "direction": ALL,
                    "component": "input",
                    "component2": "width",
                },
                "value",
            ),
        )
        def _store_grid_range_from_crop_widget(
            input_vals: List[int], width_vals: List[int]
        ) -> List[List[int]]:
            """Converts the ijk indices from one-based to zero-based and stores
            the filter range in a dcc.Store."""

            if not input_vals or not width_vals:
                return no_update
            if any(val is None for val in input_vals + width_vals):
                return no_update
            return [
                [val - 1, val + width - 2] for val, width in zip(input_vals, width_vals)
            ]


def crop_widget(
    widget_id: str,
    min_val: int,
    max_val: int,
    direction: str,
    selected_min_val: Optional[int] = None,
    selected_max_val: Optional[int] = None,
) -> html.Div:

    return html.Div(
        style={"borderBottom": "1px outset", "marginBottom": "15px"}
        if direction != GRIDDIRECTION.K
        else {},
        children=[
            html.Div(
                style={
                    "display": "grid",
                    "marginBotton": "0px",
                    "gridTemplateColumns": "2fr 1fr 8fr",
                },
                children=[
                    wcc.Label(
                        children=f"{direction} Start",
                        style={
                            "fontSize": "0.7em",
                            "fontWeight": "bold",
                            "marginRight": "5px",
                        },
                    ),
                    dcc.Input(
                        style={"width": "50px", "height": "10px"},
                        id={
                            "id": widget_id,
                            "direction": direction,
                            "component": "input",
                            "component2": "start",
                        },
                        type="number",
                        persistence=True,
                        persistence_type="session",
                        minLength=1,
                        value=selected_min_val if selected_min_val else min_val,
                        min=min_val,
                        max=max_val,
                    ),
                    wcc.Slider(
                        id={
                            "id": widget_id,
                            "direction": direction,
                            "component": "slider",
                            "component2": "start",
                        },
                        marks={val: val for val in [min_val, max_val]},
                        min=min_val,
                        max=max_val,
                        value=selected_min_val if selected_min_val else min_val,
                        step=1,
                    ),
                ],
            ),
            html.Div(
                style={
                    "display": "grid",
                    "marginTop": "0px",
                    "padding": "0px",
                    "gridTemplateColumns": "2fr 1fr 8fr",
                },
                children=[
                    wcc.Label(
                        children="Width",
                        style={
                            "fontSize": "0.7em",
                            "textAlign": "right",
                            "marginRight": "5px",
                        },
                    ),
                    dcc.Input(
                        style={"width": "50px", "height": "10px"},
                        id={
                            "id": widget_id,
                            "direction": direction,
                            "component": "input",
                            "component2": "width",
                        },
                        type="number",
                        minLength=1,
                        persistence=True,
                        persistence_type="session",
                        value=selected_max_val if selected_max_val else max_val,
                        min=1,
                        max=max_val,
                    ),
                    wcc.Slider(
                        id={
                            "id": widget_id,
                            "direction": direction,
                            "component": "slider",
                            "component2": "width",
                        },
                        min=1,
                        marks={val: "" for val in [min_val, max_val]},
                        max=max_val,
                        value=selected_max_val if selected_max_val else max_val,
                        step=1,
                    ),
                ],
            ),
        ],
    )



from typing import Any, Dict, List, Optional, Union

import webviz_core_components as wcc
from dash import Input, Output, State, callback
from dash.exceptions import PreventUpdate
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import ViewABC

from ....._figures import BarChart, ScatterPlot
from ..._reusable_view_element import GeneralViewElement
from ..._types import CorrType, DepthType, LineType
from ..._utils import FormationFigure, RftPlotterDataModel, correlate
from ._settings import Options, ParameterFilterSettings, Selections


class ParameterResponseView(ViewABC):
    class Ids(StrEnum):
        SELECTIONS = "selections"
        OPTIONS = "options"
        PARAMETER_FILTER = "parameter-filter"
        FORMATION_PLOT = "formation-plot"
        CORR_BARCHART = "corr-barchart"
        CORR_BARCHART_FIGURE = "corr-barchart-figure"
        SCATTERPLOT = "scatterplot"

    def __init__(self, datamodel: RftPlotterDataModel) -> None:
        super().__init__("Parameter Response")
        self._datamodel = datamodel
        self._parameter_df = datamodel.param_model.dataframe

        self.add_settings_group(Selections(self._datamodel), self.Ids.SELECTIONS)
        self.add_settings_group(Options(), self.Ids.OPTIONS)
        self.add_settings_group(
            ParameterFilterSettings(
                parameter_df=self._datamodel.param_model.dataframe,
                mc_ensembles=self._datamodel.param_model.mc_ensembles,
            ),
            self.Ids.PARAMETER_FILTER,
        )

        first_column = self.add_column()
        first_column.add_view_element(GeneralViewElement(), self.Ids.CORR_BARCHART)
        first_column.add_view_element(GeneralViewElement(), self.Ids.SCATTERPLOT)
        second_column = self.add_column()
        second_column.add_view_element(GeneralViewElement(), self.Ids.FORMATION_PLOT)

    def set_callbacks(self) -> None:

        corr_barchart_figure_id = self.view_element(
            self.Ids.CORR_BARCHART
        ).register_component_unique_id(self.Ids.CORR_BARCHART_FIGURE)

        @callback(
            Output(
                self.settings_group(self.Ids.SELECTIONS)
                .component_unique_id(Selections.Ids.PARAM)
                .to_string(),
                "value",
            ),
            Input(corr_barchart_figure_id, "clickData"),
            State(
                self.settings_group(self.Ids.OPTIONS)
                .component_unique_id(Options.Ids.CORRTYPE)
                .to_string(),
                "value",
            ),
            prevent_initial_call=True,
        )
        @callback_typecheck
        def _update_param_from_clickdata(
            corr_vector_clickdata: Union[None, dict],
            corrtype: CorrType,
        ) -> str:
            """Update the selected parameter from clickdata"""
            if corr_vector_clickdata is None or corrtype == CorrType.PARAM_VS_SIM:
                raise PreventUpdate
            return corr_vector_clickdata.get("points", [{}])[0].get("y")

        @callback(
            Output(
                self.settings_group(self.Ids.SELECTIONS)
                .component_unique_id(Selections.Ids.WELL)
                .to_string(),
                "value",
            ),
            Input(corr_barchart_figure_id, "clickData"),
            State(
                self.settings_group(self.Ids.OPTIONS)
                .component_unique_id(Options.Ids.CORRTYPE)
                .to_string(),
                "value",
            ),
            prevent_initial_call=True,
        )
        @callback_typecheck
        def _update_selections_from_clickdata(
            corr_vector_clickdata: Union[None, dict],
            corrtype: CorrType,
        ) -> str:
            """Update well, date and zone from clickdata"""
            if corr_vector_clickdata is None or corrtype == CorrType.SIM_VS_PARAM:
                raise PreventUpdate

            clickdata = corr_vector_clickdata.get("points", [{}])[0].get("y")
            ls_clickdata = clickdata.split()
            return ls_clickdata[0]

        @callback(
            Output(
                self.view_element(self.Ids.CORR_BARCHART)
                .component_unique_id(GeneralViewElement.Ids.CHART)
                .to_string(),
                "children",
            ),
            Output(
                self.view_element(self.Ids.SCATTERPLOT)
                .component_unique_id(GeneralViewElement.Ids.CHART)
                .to_string(),
                "children",
            ),
            Output(
                self.view_element(self.Ids.FORMATION_PLOT)
                .component_unique_id(GeneralViewElement.Ids.CHART)
                .to_string(),
                "children",
            ),
            Input(
                self.settings_group(self.Ids.SELECTIONS)
                .component_unique_id(Selections.Ids.ENSEMBLE)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.SELECTIONS)
                .component_unique_id(Selections.Ids.WELL)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.SELECTIONS)
                .component_unique_id(Selections.Ids.DATE)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.SELECTIONS)
                .component_unique_id(Selections.Ids.ZONE)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.SELECTIONS)
                .component_unique_id(Selections.Ids.PARAM)
                .to_string(),
                "value",
            ),
            Input(
                self.settings_group(self.Ids.OPTIONS)
                .component_unique_id(Options.Ids.CORRTYPE)
                .to_string(),
                "value",
            ),
            Input(
                {
                    "id": self.settings_group(self.Ids.PARAMETER_FILTER)
                    .component_unique_id(ParameterFilterSettings.Ids.PARAM_FILTER)
                    .to_string(),
                    "type": "data-store",
                },
                "data",
            ),
            Input(
                self.settings_group(self.Ids.OPTIONS)
                .component_unique_id(Options.Ids.DEPTHTYPE)
                .to_string(),
                "value",
            ),
        )
        @callback_typecheck
        # pylint: disable=too-many-locals
        def _update_paramresp_graphs(
            ensemble: str,
            well: str,
            date: str,
            zone: str,
            param: Optional[str],
            corrtype: CorrType,
            real_filter: Dict[str, List[int]],
            depthtype: DepthType,
        ) -> List[Optional[Any]]:
            """Main callback to update the graphs:
            * ranked correlations bar chart
            * response vs param scatter plot
            * formations chart RFT pressure vs depth, colored by parameter value
            """
            (
                df,
                obs,
                obs_err,
                ens_params,
                ens_rfts,
            ) = self._datamodel.create_rft_and_param_pivot_table(
                ensemble=ensemble,
                well=well,
                date=date,
                zone=zone,
                reals=real_filter[ensemble],
                keep_all_rfts=(corrtype == CorrType.PARAM_VS_SIM),
            )
            current_key = f"{well} {date} {zone}"

            if df is None:
                # This happens if the filtering criterias returns no data
                # Could f.ex happen when there are ensembles with different well names
                return ["No data matching the given filter criterias."] * 3
            if param is not None and param not in ens_params:
                # This happens if the selected parameter does not exist in the
                # selected ensemble
                return ["The selected parameter not valid for selected ensemble."] * 3
            if not ens_params:
                # This happens if there are multiple ensembles and one of the ensembles
                # doesn't have non-constant parameters.
                return ["The selected ensemble has no non-constant parameters."] * 3

            if corrtype == CorrType.SIM_VS_PARAM or param is None:
                corrseries = correlate(df[ens_params + [current_key]], current_key)
                param = param if param is not None else corrseries.abs().idxmax()
                corr_title = f"{current_key} vs parameters"
                scatter_x, scatter_y, highlight_bar = param, current_key, param

            if corrtype == CorrType.PARAM_VS_SIM:
                corrseries = correlate(df[ens_rfts + [param]], param)
                corr_title = f"{param} vs simulated RFTs"
                scatter_x, scatter_y, highlight_bar = param, current_key, current_key

            # Correlation bar chart
            corrfig = BarChart(corrseries, n_rows=15, title=corr_title, orientation="h")
            corrfig.color_bars(highlight_bar, "#007079", 0.5)
            corr_graph = wcc.Graph(
                style={"height": "40vh"},
                figure=corrfig.figure,
                id=corr_barchart_figure_id,
            )

            # Scatter plot
            scatterplot = ScatterPlot(
                df, scatter_y, scatter_x, "#007079", f"{current_key} vs {param}"
            )
            scatterplot.add_vertical_line_with_error(
                obs,
                obs_err,
                df[param].min(),
                df[param].max(),
            )
            scatter_graph = (
                wcc.Graph(
                    style={"height": "40vh"},
                    figure=scatterplot.figure,
                ),
            )

            # Formations plot
            formations_figure = FormationFigure(
                well=well,
                ertdf=self._datamodel.ertdatadf,
                enscolors=self._datamodel.enscolors,
                depthtype=depthtype,
                date=date,
                ensembles=[ensemble],
                reals=real_filter[ensemble],
                simdf=self._datamodel.simdf,
                obsdf=self._datamodel.obsdatadf,
            )

            if formations_figure.use_ertdf:
                return [
                    corr_graph,
                    scatter_graph,
                    f"Realization lines not available for depth option {depthtype}",
                ]

            if self._datamodel.formations is not None:
                formations_figure.add_formation(
                    self._datamodel.formationdf, fill_color=False
                )

            formations_figure.add_simulated_lines(LineType.REALIZATION)
            formations_figure.add_additional_observations()
            formations_figure.add_ert_observed()

            df_value_norm = self._datamodel.get_param_real_and_value_df(
                ensemble, parameter=param, normalize=True
            )
            formations_figure.color_by_param_value(df_value_norm, param)

            return [
                corr_graph,
                scatter_graph,
                wcc.Graph(
                    style={"height": "87vh"},
                    figure=formations_figure.figure,
                ),
            ]

        @callback(
            Output(
                {
                    "id": ParameterFilterSettings.Ids.PARAM_FILTER,
                    "type": "ensemble-update",
                },
                "data",
            ),
            Input(
                self.settings_group(self.Ids.SELECTIONS)
                .component_unique_id(Selections.Ids.ENSEMBLE)
                .to_string(),
                "value",
            ),
        )
        def _update_parameter_filter_selection(ensemble: str) -> List[str]:
            """Update ensemble in parameter filter"""
            return [ensemble]



import re
from enum import Enum
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
from webviz_config import WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore


class QcFlags(str, Enum):
    """Constants for use by check_swatinit"""

    FINE_EQUIL = "FINE_EQUIL"
    HC_BELOW_FWL = "HC_BELOW_FWL"
    PC_SCALED = "PC_SCALED"
    PPCWMAX = "PPCWMAX"
    SWATINIT_1 = "SWATINIT_1"
    SWL_TRUNC = "SWL_TRUNC"
    UNKNOWN = "UNKNOWN"
    WATER = "WATER"


class SwatinitQcDataModel:
    """Class keeping the data needed in the vizualisations and various
    data providing methods.
    """

    COLNAME_THRESHOLD = "HC cells above threshold (%)"
    SELECTORS = ["QC_FLAG", "SATNUM", "EQLNUM", "FIPNUM"]

    DROP_COLUMNS = [
        "Z_MIN",
        "Z_MAX",
        "GLOBAL_INDEX",
        "SWATINIT_SWAT",
        "SWATINIT_SWAT_WVOL",
        "SWL_x",
        "SWL_y",
        "Z_DATUM",
        "PRESSURE_DATUM",
    ]

    def __init__(
        self,
        webviz_settings: WebvizSettings,
        csvfile: str,
        ensemble: str = None,
        realization: Optional[int] = None,
        faultlines: Optional[Path] = None,
    ):

        self._theme = webviz_settings.theme
        self._faultlines = faultlines
        self.faultlines_df = read_csv(faultlines) if faultlines else None

        if ensemble is not None:
            if isinstance(ensemble, list):
                raise TypeError(
                    'Incorrent argument type, "ensemble" must be a string instead of a list'
                )
            if realization is None:
                raise ValueError('Incorrent arguments, "realization" must be specified')

            ens_path = webviz_settings.shared_settings["scratch_ensembles"][ensemble]
            # replace realization in string from scratch_ensemble with input realization
            ens_path = re.sub(
                "realization-[^/]", f"realization-{realization}", ens_path
            )
            self.csvfile = Path(ens_path) / csvfile
        else:
            self.csvfile = Path(csvfile)

        self.dframe = read_csv(self.csvfile)
        self.dframe.drop(columns=self.DROP_COLUMNS, errors="ignore", inplace=True)

        for col in self.SELECTORS + ["OWC", "GWC", "GOC"]:
            if col in self.dframe:
                self.dframe[col] = self.dframe[col].astype("category")

        self._initial_qc_volumes = self.compute_qc_volumes(self.dframe)

    @property
    def webviz_store(self) -> List[Tuple[Callable, List[dict]]]:
        return [
            (
                read_csv,
                [
                    {"csv_file": path}
                    for path in [self._faultlines, self.csvfile]
                    if path is not None
                ],
            )
        ]

    @property
    def colors(self) -> List[str]:
        return self._theme.plotly_theme["layout"]["colorway"]

    @property
    def qc_flag_colors(self) -> Dict[str, str]:
        """Predefined colors for the QC_FLAG column"""
        return {
            QcFlags.FINE_EQUIL.value: self.colors[8],
            QcFlags.HC_BELOW_FWL.value: self.colors[5],
            QcFlags.PC_SCALED.value: self.colors[2],
            QcFlags.PPCWMAX.value: self.colors[9],
            QcFlags.SWATINIT_1.value: self.colors[6],
            QcFlags.SWL_TRUNC.value: self.colors[3],
            QcFlags.UNKNOWN.value: self.colors[1],
            QcFlags.WATER.value: self.colors[0],
        }

    @property
    def eqlnums(self) -> List[str]:
        return sorted(list(self.dframe["EQLNUM"].unique()), key=int)

    @property
    def satnums(self) -> List[str]:
        return sorted(list(self.dframe["SATNUM"].unique()), key=int)

    @property
    def qc_flag(self) -> List[str]:
        return sorted(list(self.dframe["QC_FLAG"].unique()))

    @property
    def filters_discrete(self) -> List[str]:
        return ["QC_FLAG", "SATNUM"]

    @property
    def filters_continuous(self) -> List[str]:
        return ["Z", "PC", "SWATINIT", "PERMX", "PORO"]

    @property
    def color_by_selectors(self) -> List[str]:
        return self.SELECTORS + ["PERMX", "PORO"]

    @property
    def pc_scaling_min_max(self) -> Tuple[float, float]:
        return (self.dframe["PC_SCALING"].max(), self.dframe["PC_SCALING"].min())

    @property
    def vol_diff_total(self) -> Tuple[float, float]:
        return (
            self._initial_qc_volumes["WVOL_DIFF_PERCENT"],
            self._initial_qc_volumes["HCVOL_DIFF_PERCENT"],
        )

    def get_dataframe(
        self,
        filters: Optional[dict] = None,
        range_filters: Optional[dict] = None,
    ) -> pd.DataFrame:

        df = self.dframe.copy()
        filters = filters if filters is not None else {}
        range_filters = range_filters if range_filters is not None else {}

        for filt, value in filters.items():
            df = df[df[filt].isin(value)]

        for filt, value in range_filters.items():
            min_val, max_val = value
            df = df[(df[filt] >= min_val) & (df[filt] <= max_val) | (df[filt].isnull())]

        return df

    @staticmethod
    def resample_dataframe(dframe: pd.DataFrame, max_points: int) -> pd.DataFrame:
        """Resample a dataframe to max number of points. The sampling will be
        weighted in order to avoid removal of points that has an important qc_flag.
        Points will mostly be removed if they are flagged as "WATER" or "PC_SCALED"
        """
        if dframe.shape[0] > max_points:
            dframe = dframe.copy()
            dframe["sample_weight"] = 1
            dframe.loc[dframe["QC_FLAG"] == "WATER", "sample_weight"] = 0.1
            dframe.loc[dframe["QC_FLAG"] == "PC_SCALED", "sample_weight"] = 0.5
            return dframe.sample(max_points, weights=dframe["sample_weight"])
        return dframe

    @staticmethod
    def filter_dframe_on_depth(dframe: pd.DataFrame) -> pd.DataFrame:
        """Suggest a deep depth limit for what to plot, in order to avoid
        showing too much of a less interesting water zone
        """
        max_z = dframe["Z"].max()
        hc_dframe = dframe[dframe["SWATINIT"] < 1]
        if not hc_dframe.empty:
            lowest_hc = hc_dframe["Z"].max()
            hc_height = lowest_hc - dframe["Z"].min()
            # Suggest to visualize a water height of 10% of the hc zone:
            max_z = lowest_hc + 0.2 * hc_height

        return dframe[dframe["Z"] <= max_z]

    @staticmethod
    def compute_qc_volumes(dframe: pd.DataFrame) -> dict:
        """Compute numbers relevant for QC of saturation initialization of a
        reservoir model.
        Different volume numbers are typically related to the different QC_FLAG
        """
        qc_vols: dict = {}

        # Ensure all QCFlag categories are represented:
        for qc_cat in QcFlags:
            qc_vols[qc_cat.value] = 0.0

        # Overwrite dict values with correct figures:
        for qc_cat, qc_df in dframe.groupby("QC_FLAG"):
            qc_vols[qc_cat] = (
                (qc_df["SWAT"] - qc_df["SWATINIT"]) * qc_df["PORV"]
            ).sum()

        if "VOLUME" in dframe:
            qc_vols["VOLUME"] = dframe["VOLUME"].sum()

        qc_vols["PORV"] = dframe["PORV"].sum()
        qc_vols["SWATINIT_WVOL"] = (dframe["SWATINIT"] * dframe["PORV"]).sum()
        qc_vols["SWATINIT_HCVOL"] = qc_vols["PORV"] - qc_vols["SWATINIT_WVOL"]
        qc_vols["SWAT_WVOL"] = (dframe["SWAT"] * dframe["PORV"]).sum()
        qc_vols["SWAT_HCVOL"] = qc_vols["PORV"] - qc_vols["SWAT_WVOL"]

        # compute difference columns
        qc_vols["WVOL_DIFF"] = qc_vols["SWAT_WVOL"] - qc_vols["SWATINIT_WVOL"]
        qc_vols["WVOL_DIFF_PERCENT"] = (
            qc_vols["WVOL_DIFF"] / qc_vols["SWATINIT_WVOL"]
        ) * 100
        qc_vols["HCVOL_DIFF"] = qc_vols["SWAT_HCVOL"] - qc_vols["SWATINIT_HCVOL"]
        qc_vols["HCVOL_DIFF_PERCENT"] = (
            ((qc_vols["HCVOL_DIFF"] / qc_vols["SWATINIT_HCVOL"]) * 100)
            if qc_vols["HCVOL_DIFF"] != 0.0
            else 0
        )
        qc_vols["EQLNUMS"] = sorted(dframe["EQLNUM"].unique())
        qc_vols["SATNUMS"] = sorted(dframe["SATNUM"].unique())

        return qc_vols

    def create_colormap(self, color_by: str) -> dict:
        """Create a colormap to ensure that the subplot and the mapfigure
        has the same color for the same unique value. If 'QC_FLAG' is used as
        color column, predefined colors are used.
        """

        return (
            dict(zip(self.dframe[color_by].unique(), self.colors * 10))
            if color_by != "QC_FLAG"
            else self.qc_flag_colors
        )

    def get_max_pc_info_and_percent_for_data_matching_condition(
        self,
        dframe: pd.DataFrame,
        condition: Optional[int],
        groupby_eqlnum: bool = True,
    ) -> pd.DataFrame:
        def get_percent_of_match(df: pd.DataFrame, condition: Optional[int]) -> float:
            df = df[df["QC_FLAG"] == "PC_SCALED"]
            if condition is None or df.empty:
                return np.nan
            return (len(df[df["PC_SCALING"] >= condition]) / len(df)) * 100

        groupby = ["SATNUM"] if not groupby_eqlnum else ["EQLNUM", "SATNUM"]
        df_group = dframe.groupby(groupby)
        df = df_group.max()[["PCOW_MAX", "PPCW", "PC_SCALING"]].round(6)
        df[self.COLNAME_THRESHOLD] = df_group.apply(
            lambda x: get_percent_of_match(x, condition)
        )
        return df.reset_index().sort_values(groupby, key=lambda col: col.astype(int))

    def table_data_qc_vol_overview(self) -> tuple:
        """Return data and columns for dash_table showing overview of qc volumes"""

        skip_if_zero = [QcFlags.UNKNOWN.value, QcFlags.WATER.value]
        column_order = [
            "",
            "Response",
            "Water Volume Diff",
            "HC Volume Diff",
            "Water Volume Mrm3",
            "HC Volume Mrm3",
        ]
        qc_vols = self._initial_qc_volumes

        table_data = []
        # First report the SWATINIT volumes
        table_data.append(
            {
                "Response": "SWATINIT",
                "Water Volume Mrm3": f"{qc_vols['SWATINIT_WVOL']/1e6:>10.3f}",
                "HC Volume Mrm3": f" {qc_vols['SWATINIT_HCVOL']/1e6:>8.3f}",
            }
        )
        # Then report the volume change per QC_FLAG
        for key in [x.value for x in QcFlags]:
            if key in skip_if_zero and np.isclose(qc_vols[key], 0, atol=1):
                # Tolerance is 1 rm3, which is small in relevant contexts.
                continue
            table_data.append(
                {
                    "": "+",
                    "Response": key,
                    "Water Volume Mrm3": f"{qc_vols[key]/1e6:>10.3f}",
                    "Water Volume Diff": f"{qc_vols[key]/qc_vols['SWATINIT_WVOL']*100:>3.2f} %",
                    "HC Volume Diff": f"{-qc_vols[key]/qc_vols['SWATINIT_HCVOL']*100:>3.2f} %"
                    if qc_vols["SWATINIT_HCVOL"] > 0
                    else "0.00 %",
                }
            )
        # Last report the SWAT volumes and change from SWATINIT
        table_data.append(
            {
                "": "=",
                "Response": "SWAT",
                "Water Volume Mrm3": f"{qc_vols['SWAT_WVOL']/1e6:>10.3f}",
                "Water Volume Diff": f"{qc_vols['WVOL_DIFF_PERCENT']:>3.2f} %",
                "HC Volume Diff": f"{qc_vols['HCVOL_DIFF_PERCENT']:>3.2f} %",
                "HC Volume Mrm3": f"{qc_vols['SWAT_HCVOL']/1e6:>8.3f}",
            }
        )
        return table_data, [{"name": i, "id": i} for i in column_order]


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file: str) -> pd.DataFrame:
    return pd.read_csv(csv_file)



import time


class PerfTimer:
    def __init__(self) -> None:
        self._start_s = time.perf_counter()
        self._lap_s = self._start_s

    def elapsed_s(self) -> float:
        """Get the time elapsed since the start, in seconds"""
        return time.perf_counter() - self._start_s

    def elapsed_ms(self) -> int:
        """Get the time elapsed since the start, in milliseconds"""
        return int(1000 * self.elapsed_s())

    def lap_s(self) -> float:
        """Get elapsed time since last lap, in seconds"""
        time_now = time.perf_counter()
        elapsed = time_now - self._lap_s
        self._lap_s = time_now
        return elapsed

    def lap_ms(self) -> int:
        """Get elapsed time since last lap, in milliseconds"""
        return int(1000 * self.lap_s())
